from src.classes.PopupMessage import PopupMessage

import pygame
import sys

c1 = {
    'black': (0,0,0),
    'white': (255,255,255),
    'cyan': (0,255,255),
    'magenta': (255,0,255),
    'yellow': (255,255,0),
    'red': (255,0,0),
    'green': (0,255,0),
    'blue': (0,0,255)
}

pygame.init()
screen = pygame.display.set_mode((500,500))
fpsclock = pygame.time.Clock()
fps = 60

def circle(xy, color, radius=3, width=0):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.circle(screen, color, xy, radius, width)

def line(xy1, xy2, color, width=1):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.line(screen, color, xy1, xy2, width)

def rect(r, color, width=0):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.rect(screen, color, r, width)

center = (250, 250)

msg = PopupMessage((250,  320), (250, 180), screen, 32, "center", fh=40, xad=10, bw=1)

while True:
    events = pygame.event.get()
    for event in events:
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            pygame.quit()
            sys.exit()
        if event.type == pygame.KEYDOWN and event.key == pygame.K_SPACE:
            print("showing test message")
            msg.show("Test message", 4)
    
    msg.update()

    screen.fill((20,20,200))

    line((75, 250), (425, 250), (0,0,0), 3)
    line((210, 200), (290, 200), (0,0,0), 3)
    line((210, 300), (290, 300), (0,0,0), 3)

    msg.draw()


    pygame.display.flip()
    fpsclock.tick(fps)


class OneClickManager:
    def __init__(self):
        self.clicked = False
        self.hovering = False

    def update(self):
        self.clicked = False
        self.hovering = False

    def get_clicked(self):
        return self.clicked

    def get_hovering(self):
        return self.hovering
    
    def set_clicked(self, status=True):
        self.clicked = bool(status)

    def set_hovering(self, status=True):
        self.hovering = bool(status)


from .Label import Label
from .Button import Button
from .Entry import Entry


class CustomTemplate:
    def __init__(self, **kwargs):
        self.__kwargs = kwargs

    def get(self, key):
        return self.__kwargs.get(key)

    def label(self, **kwargs) -> Label:
        local_kwargs = self.__kwargs.copy()
        for k, v in kwargs.items():
            local_kwargs[k] = v

        assert "surface" in local_kwargs.keys(), "<surface> has to be set"
        assert "text" in local_kwargs.keys(), "<text> has to be set"
        assert "size" in local_kwargs.keys(), "<size> has to be set"
        assert "xy" in local_kwargs.keys(), "<xy> has to be set"
        assert "anchor" in local_kwargs.keys(), "<anchor> has to be set"

        return Label(
            local_kwargs.pop("surface"),
            local_kwargs.pop("text"),
            local_kwargs.pop("size"),
            local_kwargs.pop("xy"),
            local_kwargs.pop("anchor"),
            **local_kwargs)
    
    def button(self, **kwargs) -> Button:
        local_kwargs = self.__kwargs.copy()
        for k, v in kwargs.items():
            local_kwargs[k] = v

        assert "surface" in local_kwargs.keys(), "<surface> has to be set"
        assert "text" in local_kwargs.keys(), "<text> has to be set"
        assert "size" in local_kwargs.keys(), "<size> has to be set"
        assert "xy" in local_kwargs.keys(), "<xy> has to be set"
        assert "anchor" in local_kwargs.keys(), "<anchor> has to be set"

        return Button(
            local_kwargs.pop("surface"),
            local_kwargs.pop("text"),
            local_kwargs.pop("size"),
            local_kwargs.pop("xy"),
            local_kwargs.pop("anchor"),
            **local_kwargs)
    
    def entry(self, **kwargs) -> Entry:
        local_kwargs = self.__kwargs.copy()
        for k, v in kwargs.items():
            local_kwargs[k] = v

        assert "surface" in local_kwargs.keys(), "<surface> has to be set"
        assert "text" in local_kwargs.keys(), "<text> has to be set"
        assert "size" in local_kwargs.keys(), "<size> has to be set"
        assert "xy" in local_kwargs.keys(), "<xy> has to be set"
        assert "anchor" in local_kwargs.keys(), "<anchor> has to be set"

        return Entry(
            local_kwargs.pop("surface"),
            local_kwargs.pop("text"),
            local_kwargs.pop("size"),
            local_kwargs.pop("xy"),
            local_kwargs.pop("anchor"),
            **local_kwargs)



import math
import pygame
import sys
import os
from .Label import Label
from .Button import Button

class Function:
    def __init__(self, min_: int = 0, max_: int = 1, function_variable: str = "x"):
        """ scales everything to a function between min_ and max_ (on the y axis) """
        self.__min = min_
        self.__max = max_
        self.__function_variable = function_variable
        self.__length = 0
        self.__funcs = [
            # every func except last:
            # [from (including), to (excluding), f_from, f_to, func, x_dist, y_dist]
            # last func:
            # [from (including), to (including), f_from, f_to, func, x_dist, y_dist]
        ]

        assert isinstance(function_variable, str) and function_variable.isalpha(), \
            f"invalid argument for 'function_variable': {function_variable}"

        self.__val_before = None
        self.__val_after = None

    def add_const(self, constant: float, length: float = 1):
        """ add a constant value """
        assert length > 0, f"<length> as to be greater than 0"
        assert isinstance(constant, (float, int)), f"invalid argument for 'constant': {constant}"
        self.__funcs.append(
            {
                # i_... = internal_... -> regarding the internal function that is being constructed
                # f_... = function_... -> regarding the function that is given by <func>
                "value": constant,
                "i_from": self.__length,
                "i_to": self.__length + length,
                "is_constant": True,
            }
        )
        self.__length += length

    def add_func(self, func: str, f_from: float, f_to: float, length: float = 1, start_at="min"):
        """ use the <function_variable> (by default 'x'), eg. '-x**2'; use 'math.' for the math library;
        <start_at> can be either 'min' or 'max' """

        assert isinstance(eval(func, {self.__function_variable: 0, "math": math}), int) or \
            isinstance(eval(func, {self.__function_variable: 0, "math": math}), float), \
            f"invalid argument for 'func': {func} | error: does not return integer or float"
        assert f_from < f_to, f"<f_from> can not be greater than or equal to <f_to>"
        assert length > 0, f"<length> as to be greater than 0"
        assert start_at in ("min", "max"), f"invalid argument for 'start_at': {start_at}"
        # assert (start_at in ["min", "max"] or isinstance(start_at, int) or isinstance(start_at, float)), \
        #     f"invalid argument for 'start_at': {start_at}"
        self.__funcs.append(
            {
                # i_... = internal_... -> regarding the internal function that is being constructed
                # f_... = function_... -> regarding the function that is given by <func>
                "func": func,
                "i_from": self.__length,
                "i_to": self.__length + length,
                "i_xdist": length,
                "i_ydist": self.__max - self.__min,
                "f_from": f_from,
                "f_to": f_to,
                "f_xdist": f_to - f_from,
                "f_ydist": abs(eval(func, {"x": f_to, "math": math}) - eval(func, {"x": f_from, "math": math})),
                "start_at": start_at,
                "is_constant": False,
            }
        )
        self.__length += length

    def __round_half_up(self, n, decimals):
        if isinstance(n, int):
            return n
        multiplier = 10 ** decimals
        if decimals == 0:
            return int(math.floor(n*multiplier + 0.5) / multiplier)
        else:
            return math.floor(n*multiplier + 0.5) / multiplier

    def get(self, x, decimals=None):
        """ return the value of the function at <x> """

        if not 0 <= x <= self.__length:
            if self.__val_before is not None and x < 0:
                return self.__val_before
            elif self.__val_after is not None and x > self.__length:
                return self.__val_after
            else:
                raise ValueError(f"<x> is out of range! Function length: {self.__length}")

        # finding the correct function (setting to 'f')
        for count, func in enumerate(self.__funcs):

            # if last function
            if count == len(self.__funcs)-1:
                if func["i_from"] <= x <= func["i_to"]:
                    f = func
                    break

            # every function but the last
            else:
                if func["i_from"] <= x < func["i_to"]:
                    f = func
                    break

        # now we have 'f' as the correct function
        if f["is_constant"] == False:
            actual_x = x - func["i_from"]
            percent = actual_x / func["i_xdist"]

            new_x_dist = func["f_xdist"]
            filling = percent * new_x_dist
            f_x = func["f_from"] + filling

            y_val = eval(f["func"],
                         {self.__function_variable: f_x, "math": math})

            y_min = eval(f["func"],
                         {self.__function_variable: func["f_from"], "math": math})
            y_max = eval(f["func"],
                         {self.__function_variable: func["f_to"], "math": math})
            g = abs(y_max - y_min)
            a = y_val - y_min
            factor = a/g

            if f["start_at"] == "min":
                start = self.__min
                end = self.__max
            elif f["start_at"] == "max":
                start = self.__max
                end = self.__min
            # else:
            #     start = f["start_at"]

            filling = factor * abs(end - start)
            value = start + filling

        elif f["is_constant"] == True:
            value = f["value"]

        if decimals != None:
            return self.__round_half_up(value, decimals)
        else:
            return value

    def set_outer_values(self, value_before: float, value_after: float):
        assert isinstance(value_before, int) or isinstance(
            value_before, float), f"invalid argument for 'value_before': {value_before}"
        assert isinstance(value_after, int) or isinstance(
            value_after, float), f"invalid argument for 'value_after': {value_after}"
        self.__val_before = value_before
        self.__val_after = value_after

    def show(self, show_every_value=False):
        """ display the current function in a window """

        if self.__length == 0:
            raise Exception("Can not draw function because the length is zero.")

        # creating the dots # ! improvement needed # TODO
        dots = []
        dots_count = 200
        step = self.__length / dots_count
        x = 0
        while x <= self.__length:
            dots.append(
                (x, self.get(x))
            )
            x += step

        d_min = min([d[1] for d in dots])
        d_max = max([d[1] for d in dots])

        if show_every_value:
            MIN = d_min
            MAX = d_max
        else:
            MIN = self.__min
            MAX = self.__max

        pygame.init()
        screen = pygame.display.set_mode((900, 500))
        fpsclock = pygame.time.Clock()
        fps = 60

        col = [160 for i in range(3)]

        l_y_max = Label(screen, self.__round_half_up(MAX, 2), 20, (50, 50), "midright", xad=10, tc=col)
        l_y_min = Label(screen, self.__round_half_up(MIN, 2), 20, (50, 450), "midright", xad=10, tc=col)
        l_y_max.draw()
        l_y_min.draw()

        # if show_every_value: also draw self.__min and self.__max marks
        if show_every_value:
            l_mark_max_percent = (self.__max - MIN) / (MAX - MIN)
            l_mark_max_height = 450 - 400*l_mark_max_percent
            if self.__max != MAX:
                l_y_max_mark = Label(
                    screen, self.__round_half_up(self.__max, 3),
                    20, (50, l_mark_max_height),
                    "midright", xad=10, tc=col)
                l_y_max_mark.draw()

            l_mark_min_percent = (self.__min - MIN) / (MAX - MIN)
            l_mark_min_height = 450 - 400*l_mark_min_percent
            if self.__min != MIN:
                l_y_min_mark = Label(
                    screen, self.__round_half_up(self.__min, 3),
                    20, (50, l_mark_min_height),
                    "midright", xad=10, tc=col)
                l_y_min_mark.draw()

        l_x_min = Label(screen, 0, 20, (50, 450), "midtop", yad=10, tc=col)
        l_x_max = Label(screen, self.__length, 20, (850, 450), "midtop", yad=10, tc=col)
        l_x_min.draw()
        l_x_max.draw()

        # drawing the lines
        pygame.draw.line(screen, col, (50, 50), (50, 450), 1)
        if show_every_value:
            pygame.draw.line(screen, (255, 0, 0), (50, l_mark_max_height), (850, l_mark_max_height), 1)
            pygame.draw.line(screen, (255, 0, 0), (50, l_mark_min_height), (850, l_mark_min_height), 1)
        else:
            pygame.draw.line(screen, (255, 0, 0), (50, 450), (850, 450), 1)
            pygame.draw.line(screen, (255, 0, 0), (50, 50), (850, 50), 1)

        # drawing the dots
        for dot in dots:
            min_percent = (self.__min - MIN) / (MAX - MIN)
            min_height = 400*min_percent
            pygame.draw.circle(
                screen, (0, 0, 255),
                (50 + dot[0] * (800 / self.__length),
                 450 - min_height - dot[1] * (400 / (MAX - MIN))),
                2)

        # incr_size_img = pygame.image.load(r"images\increase_size.png")
        incr_size_img = pygame.image.load(os.path.join(os.path.split(os.path.dirname(__file__))[0], "images", "increase_size.png"))
        incr_size_img.set_colorkey((255, 255, 255))
        # decr_size_img = pygame.image.load(r"images\decrease_size.png")
        decr_size_img = pygame.image.load(os.path.join(os.path.split(os.path.dirname(__file__))[0], "images", "decrease_size.png"))
        decr_size_img.set_colorkey((255, 255, 255))

        if show_every_value:
            button = Button(screen, "", 10, (22, 22), "center", fd=(32, 32), img=decr_size_img)
        else:
            button = Button(screen, "", 10, (22, 22), "center", fd=(32, 32), img=incr_size_img)
        button.draw()

        run = True
        while run:
            event_list = pygame.event.get()
            for event in event_list:
                if event.type == pygame.QUIT:
                    pygame.quit()
                    sys.exit()
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        run = False

            if button.update(event_list):
                self.show(not show_every_value)

            pygame.display.flip()
            fpsclock.tick(fps)

    def reset(self):
        """ deletes all previously added functions """
        self.__length = 0
        self.__funcs = []

    def example_function(self):
        """ show the example function """
        self.__min = 0
        self.__max = 1
        self.__function_variable = "x"
        self.__length = 0
        self.__funcs = []
        self.add_func("(-math.sin(x*3))/math.cos(x/2)", 0.5, 4.5, 10)
        print("Function: (-math.sin(x*3))/math.cos(x/2)")
        self.show()



import ctypes
import math
import os

import pygame


def higher_resolution(boolean: bool = True):
    """
    Most Windows PCs (automatically) use scaling to 150% which makes many applications appear
    blurry. To avoid this effect, this function can be called to increase the resolution on
    affected displays without having to manually change the scaling in the computers settings.
    """
    ctypes.windll.shcore.SetProcessDpiAwareness(bool(boolean))


def scale_image(image, factor):
    """
    Changes the size of an image while keeping its proportions.
    """
    return pygame.transform.scale(
        image, (int(image.get_width() * factor), int(image.get_height() * factor))
    )


def rotate_in_place(rect, image, degrees):
    """
    ONLY WORKS WITH SPRITES

    Usage:
    self.rect, self.image = rotate_in_place(self.rect, self.image)
    """
    old_rect_center = rect.center
    new_image = pygame.transform.rotate(image, degrees)
    new_rect = new_image.get_rect()
    new_rect.center = old_rect_center
    return new_rect, new_image


def create_animation(image_ids: list, frames_per_image: list):
    """
    create a dict with image_ids and corresponding
    images before using this function.

    If usage unclear, refer to:
    marcel-python-programs\PYGAME_GAMES\pve\mini_warfare\data\player.py
    """
    assert len(image_ids) == len(
        frames_per_image
    ), "received lists of different lengths"
    seq = []
    for image_id, count in zip(image_ids, frames_per_image):
        for i in range(count):
            seq.append(image_id)
    return seq


def get_distance(xy1: tuple, xy2: tuple, digits_after_comma: int = None) -> float:
    """returns the distance between xy1 and xy2"""
    x1, y1 = xy1[0], xy1[1]
    x2, y2 = xy2[0], xy2[1]
    if digits_after_comma == None:
        return math.sqrt(((x2 - x1) ** 2) + ((y2 - y1) ** 2))
    else:
        return round_half_up(
            math.sqrt(((x2 - x1) ** 2) + ((y2 - y1) ** 2)), digits_after_comma
        )


def get_distance_squared(xy1: tuple, xy2: tuple):
    return (xy1[0] - xy2[0]) ** 2 + (xy1[1] - xy2[1]) ** 2


def get_angle_between_points(xy1, xy2):
    vect = pygame.math.Vector2(xy2[0] - xy1[0], xy2[1] - xy1[1]).normalize()
    cos = round_half_up(vect.y, 2)
    deg_cos = int(math.degrees(math.acos(vect.x)))
    if cos <= 0:
        return deg_cos
    else:
        return 180 + (180 - deg_cos)


def get_normalized_vector_between_points(xy1, xy2):
    vect = pygame.math.Vector2(xy2[0] - xy1[0], xy2[1] - xy1[1])
    if vect.length() > 0:
        vect.normalize()
    return vect


def get_normalized_vector_from_angle(angle):
    vect = pygame.Vector2(
        round_half_up(math.cos(math.radians(angle)), 3),
        round_half_up(math.sin(math.radians(angle)), 3),
    )
    return vect.normalize()


def round_half_up(n, decimals=0):
    if isinstance(n, int):
        return n
    multiplier = 10**decimals
    if decimals == 0:
        return int(math.floor(n * multiplier + 0.5) / multiplier)
    else:
        return math.floor(n * multiplier + 0.5) / multiplier


def vect_sum(a: tuple, b: tuple):
    if len(a) != len(b):
        raise ValueError("Received vectors of different dimensions.")
    vect = []
    for num1, num2 in zip(a, b):
        vect.append(num1 + num2)
    return tuple(vect)


def vs(a: tuple, b: tuple):
    """vect_sum(...)"""
    return vect_sum(a, b)


def invert(a: tuple):
    """inverts a vector"""
    return (-a[0], -a[1])


def inv(a: tuple):
    """invert(...)"""
    return invert(a)


def load_image(
    path, size: tuple[int, int] = None, colorkey: tuple[int, int, int] = None
):
    """example for <path>: data/images/image1"""
    assert type(path) == str, f"invalid argument for 'path': {path}"
    if "/" in path:
        splitted = path.split("/")
        path = os.path.join(*splitted)

    try:
        img = pygame.image.load(path)
    except:
        raise Exception("image '" + str(path) + "' not found")

    if size != None:
        # assertions
        assert type(size) in [tuple, list], f"invalid argument for 'size': {size}"
        assert len(size) == 2, f"invalid argument for 'size': {size}"
        for item in size:
            assert type(item) == int, f"invalid argument for 'size': {size}"
            assert item > 0, f"invalid argument for 'size': {size}"

        img = pygame.transform.scale(img, size)

    if colorkey != None:
        # assertions
        assert type(colorkey) in [
            tuple,
            list,
        ], f"invalid argument for 'colorkey': {colorkey}"
        assert len(colorkey) == 3, f"invalid argument for 'colorkey': {colorkey}"
        for item in colorkey:
            assert type(item) == int, f"invalid argument for 'colorkey': {colorkey}"
            assert 0 <= item <= 255, f"invalid argument for 'colorkey': {colorkey}"

        img.set_colorkey(colorkey)

    return img


def draw_rect_alpha(
    surface,
    color,
    rect,
    width: int = 0,
    border_radius: int = -1,
    border_top_left_radius: int = -1,
    border_top_right_radius: int = -1,
    border_bottom_left_radius: int = -1,
    border_bottom_right_radius: int = -1,
):
    shape_surf = pygame.Surface(pygame.Rect(rect).size, pygame.SRCALPHA)
    pygame.draw.rect(
        shape_surf,
        color,
        shape_surf.get_rect(),
        width,
        border_radius,
        border_top_left_radius,
        border_top_right_radius,
        border_bottom_left_radius,
        border_bottom_right_radius,
    )
    surface.blit(shape_surf, rect)


def draw_circle_alpha(
    surface,
    color,
    center,
    radius,
    width: int = 0,
    draw_top_right: bool = None,
    draw_top_left: bool = None,
    draw_bottom_left: bool = None,
    draw_bottom_right: bool = None,
):
    target_rect = pygame.Rect(center, (0, 0)).inflate((radius * 2, radius * 2))
    shape_surf = pygame.Surface(target_rect.size, pygame.SRCALPHA)
    pygame.draw.circle(
        shape_surf,
        color,
        (radius, radius),
        radius,
        width,
        draw_top_right,
        draw_top_left,
        draw_bottom_left,
        draw_bottom_right,
    )
    surface.blit(shape_surf, target_rect)


def draw_polygon_alpha(surface, color, points):
    lx, ly = zip(*points)
    min_x, min_y, max_x, max_y = min(lx), min(ly), max(lx), max(ly)
    target_rect = pygame.Rect(min_x, min_y, max_x - min_x, max_y - min_y)
    shape_surf = pygame.Surface(target_rect.size, pygame.SRCALPHA)
    pygame.draw.polygon(shape_surf, color, [(x - min_x, y - min_y) for x, y in points])
    surface.blit(shape_surf, target_rect)


def ask_filename():
    """temporary method to get filename"""
    import tkinter.filedialog

    top = tkinter.Tk()
    top.withdraw()  # hide window
    file_name = tkinter.filedialog.askopenfilename(parent=top)
    top.destroy()
    return file_name


def ask_directory():
    """temporary method to get foldername"""
    import tkinter.filedialog

    top = tkinter.Tk()
    top.withdraw()  # hide window
    dir_name = tkinter.filedialog.askdirectory(parent=top)
    top.destroy()
    return dir_name


def get_fonts():
    return [
        f.removesuffix(".ttf")
        for f in os.listdir(os.path.join(os.path.dirname(__file__), "..", "fonts"))
    ]



class Collisions:
    @staticmethod
    def circle_circle(c1_center: tuple, c1_radius: float, c2_center: tuple, c2_radius: float) -> bool:
        return (c2_center[0] - c1_center[0]) ** 2 + (c2_center[1] - c1_center[1]) ** 2 <= (c1_radius + c2_radius) ** 2

    @staticmethod
    def rect_rect(rect1: tuple, rect2: tuple) -> bool:
        """ expects rects in the form of (x, y, width, height) """
        x1, y1, w1, h1 = rect1
        x2, y2, w2, h2 = rect2
        if x1 + w1 < x2 or x2 + w2 < x1:
            return False
        if y1 + h1 < y2 or y2 + h2 < y1:
            return False
        return True

    @staticmethod
    def circle_rect(circle_center: tuple, circle_radius: float, rect: tuple) -> bool:
        cx, cy = circle_center
        rx, ry, rw, rh = rect

        # Find the closest point to the circle within the rectangle
        closest_x = min(max(cx, rx), rx + rw)
        closest_y = min(max(cy, ry), ry + rh)

        # Calculate the distance between the circle's center and this closest point
        distance_x = cx - closest_x
        distance_y = cy - closest_y

        # If the distance is less than or equal to the circle's radius, there's an intersection
        return distance_x ** 2 + distance_y ** 2 <= circle_radius ** 2

    @staticmethod
    def line_circle(line_pos1: tuple, line_pos2: tuple, circle_center: tuple, circle_radius: float) -> bool:
        x1, y1 = line_pos1
        x2, y2 = line_pos2
        cx, cy = circle_center

        # Compute the vector from line start to circle center
        ac = (cx - x1, cy - y1)
        ab = (x2 - x1, y2 - y1)
        ab_len_squared = ab[0] ** 2 + ab[1] ** 2

        # Project ac onto ab, computing the point on ab closest to circle center
        projection = (ac[0] * ab[0] + ac[1] * ab[1]) / ab_len_squared
        closest_x = x1 + projection * ab[0]
        closest_y = y1 + projection * ab[1]

        # Clamp closest point to the segment
        if projection < 0:
            closest_x, closest_y = x1, y1
        elif projection > 1:
            closest_x, closest_y = x2, y2

        # Calculate the distance from the closest point to the circle center
        distance_x = closest_x - cx
        distance_y = closest_y - cy

        # Check if the distance is less than or equal to the circle's radius
        return distance_x ** 2 + distance_y ** 2 <= circle_radius ** 2

    @staticmethod
    def line_rect(line_pos1: tuple, line_pos2: tuple, rect: tuple) -> bool:
        rx, ry, rw, rh = rect
        rect_lines = [
            ((rx, ry), (rx + rw, ry)),
            ((rx + rw, ry), (rx + rw, ry + rh)),
            ((rx + rw, ry + rh), (rx, ry + rh)),
            ((rx, ry + rh), (rx, ry)),
        ]

        for rl1, rl2 in rect_lines:
            if Collisions.line_line(line_pos1, line_pos2, rl1, rl2):
                return True

        # Check if the line starts or ends inside the rectangle
        x1, y1 = line_pos1
        x2, y2 = line_pos2
        if (rx <= x1 <= rx + rw and ry <= y1 <= ry + rh) or (rx <= x2 <= rx + rw and ry <= y2 <= ry + rh):
            return True

        return False

    @staticmethod
    def polygon_circle(polygon: tuple, circle_center: tuple, circle_radius: float) -> bool:
        cx, cy = circle_center

        # Check if any of the polygon's vertices are within the circle
        for px, py in polygon:
            if (px - cx) ** 2 + (py - cy) ** 2 <= circle_radius ** 2:
                return True

        # Check if any of the polygon's edges intersect the circle
        for i in range(len(polygon)):
            p1 = polygon[i]
            p2 = polygon[(i + 1) % len(polygon)]
            if Collisions.line_circle(p1, p2, circle_center, circle_radius):
                return True

        return False

    @staticmethod
    def polygon_rect(polygon: tuple, rect: tuple) -> bool:
        rx, ry, rw, rh = rect

        # Check if any of the polygon's vertices are within the rectangle
        for px, py in polygon:
            if rx <= px <= rx + rw and ry <= py <= ry + rh:
                return True

        # Check if any of the polygon's edges intersect the rectangle
        for i in range(len(polygon)):
            p1 = polygon[i]
            p2 = polygon[(i + 1) % len(polygon)]
            if Collisions.line_rect(p1, p2, rect):
                return True

        return False

    @staticmethod
    def polygon_line(polygon: tuple, line_pos1: tuple, line_pos2: tuple) -> bool:
        for i in range(len(polygon)):
            p1 = polygon[i]
            p2 = polygon[(i + 1) % len(polygon)]
            if Collisions.line_line(p1, p2, line_pos1, line_pos2):
                return True
        return False

    @staticmethod
    def polygon_polygon(polygon1: tuple, polygon2: tuple) -> bool:
        # Check if any of the edges of polygon1 intersect any of the edges of polygon2
        for i in range(len(polygon1)):
            p1 = polygon1[i]
            p2 = polygon1[(i + 1) % len(polygon1)]
            if Collisions.polygon_line(polygon2, p1, p2):
                return True

        # Check if polygon2 is entirely inside polygon1 or vice versa
        if Collisions._point_in_polygon(polygon1[0], polygon2) or Collisions._point_in_polygon(polygon2[0], polygon1):
            return True

        return False

    @staticmethod
    def _point_in_polygon(point: tuple, polygon: tuple) -> bool:
        x, y = point
        inside = False
        for i in range(len(polygon)):
            xi, yi = polygon[i]
            xj, yj = polygon[(i + 1) % len(polygon)]
            intersect = ((yi > y) != (yj > y)) and (x < (xj - xi) * (y - yi) / (yj - yi) + xi)
            if intersect:
                inside = not inside
        return inside

    @staticmethod
    def line_line(line1_pos1: tuple, line1_pos2: tuple, line2_pos1: tuple, line2_pos2: tuple) -> bool:
        def ccw(A, B, C):
            return (C[1] - A[1]) * (B[0] - A[0]) > (B[1] - A[1]) * (C[0] - A[0])

        A, B = line1_pos1, line1_pos2
        C, D = line2_pos1, line2_pos2
        return ccw(A, C, D) != ccw(B, C, D) and ccw(A, B, C) != ccw(A, B, D)



import pygame
import time
from .Label import Label
from .Function import Function

class PopupMessage(Label):
    def __init__(self, passive_xy, active_xy, surface: pygame.Surface, size: int, anchor="center", **kwargs):

        # passive_xy #
        self.__passive_xy = passive_xy
        # assertion #
        assert isinstance(self.__passive_xy, (tuple, list)), f"invalid argument for 'passive_xy': {self.__passive_xy}"
        assert len(self.__passive_xy) == 2, f"invalid argument for 'passive_xy': {self.__passive_xy}"
        
        # active_xy #
        self.__active_xy = active_xy
        # assertion #
        assert isinstance(self.__active_xy, (tuple, list)), f"invalid argument for 'active_xy': {self.__active_xy}"
        assert len(self.__active_xy) == 2, f"invalid argument for 'active_xy': {self.__active_xy}"

        assert self.__active_xy != self.__passive_xy, "passive_xy and active_xy have to be different"

        super().__init__(surface, "", size, passive_xy, anchor, **kwargs)

        self.time = time.time()

        self.__f = Function()
        self.__f.add_const(0)
        self.__f.set_outer_values(0,0)

        self.__dx, self.__dy = self.__active_xy[0] - self.__passive_xy[0], self.__active_xy[1] - self.__passive_xy[1]

    def update(self):
        x = abs(self.time - time.time())
        dx = self.__dx * self.__f.get(x)
        dy = self.__dy * self.__f.get(x)
        self.update_pos((self.__passive_xy[0] + dx, self.__passive_xy[1] + dy))

    def show(self, text, seconds=4):
        self.__f.reset()
        self.__f.add_func("-x**2", -1, 0.3, 0.5)
        self.__f.add_const(1, seconds)
        self.__f.add_func("-x**2", -0.3, 1, 0.5, "max")
        self.__f.set_outer_values(0,0)

        self.update_text(str(text))
        self.time = time.time()

    def update_passive_xy(self, passive_xy):
        # passive_xy #
        self.__passive_xy = passive_xy
        # assertion #
        assert isinstance(self.__passive_xy, (tuple, list)), f"invalid argument for 'passive_xy': {self.__passive_xy}"
        assert len(self.__passive_xy) == 2, f"invalid argument for 'passive_xy': {self.__passive_xy}"
        self.__dx, self.__dy = self.__active_xy[0] - self.__passive_xy[0], self.__active_xy[1] - self.__passive_xy[1]

    def update_active_xy(self, active_xy):
        # active_xy #
        self.__active_xy = active_xy
        # assertion #
        assert isinstance(self.__active_xy, (tuple, list)), f"invalid argument for 'active_xy': {self.__active_xy}"
        assert len(self.__active_xy) == 2, f"invalid argument for 'active_xy': {self.__active_xy}"
        self.__dx, self.__dy = self.__active_xy[0] - self.__passive_xy[0], self.__active_xy[1] - self.__passive_xy[1]



from src.classes.Entry import Entry

import pygame
import sys

pygame.init()
screen = pygame.display.set_mode((600,400), display=0)
fpsclock = pygame.time.Clock()
fps = 60

e = Entry(screen, "", 32, (screen.get_width()//2, screen.get_height()//2), bw=1, fd=(400,60), twe="empty", ast=1)
e.set_state(1)

old_e = ""

run = True
while run:
    event_list = pygame.event.get()
    for event in event_list:
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                run = False

    e.update(event_list)
    if e.get_state() == 1:
        e.update_colors(bordercolor=(0,255,0))
    else:
        e.update_colors(bordercolor=(0,0,0))

    if e.get() != old_e:
        print("Changed to: " + e.get())
        old_e = e.get()
    
    screen.fill((64,64,128))
    e.draw()
    
    pygame.display.flip()
    fpsclock.tick(fps)


import pygame
import math
from enum import Enum
from .Functions import round_half_up

class PS(Enum):
    cross = 0
    circle = 1
    square = 2
    triangle = 3
    share = 4
    ps = 5
    options = 6
    l1 = 9
    l3 = 7
    r1 = 10
    r3 = 8
    arrow_up = 11
    arrow_down = 12
    arrow_left = 13
    arrow_right = 14
    touch_pad = 15

class PlayStationController:
    def __init__(self, joystick_num, threshold: float = 0.05):
        """
        Makes handling of controller input easier.
        """
        self.joystick_num = joystick_num
        assert threshold >= 0, f"threshold must be at least 0 (currently {threshold})"
        assert threshold < 1, f"threshold must be less than 1 (currently {threshold})"
        self.threshold = threshold
        pygame.init()
        self.js = pygame.joystick.Joystick(joystick_num)
        self.js.init()
        self.pressed = {}

        self.update()

    def update(self):
        keys = self.get_pressed()
        self.cross = True if keys.get(PS.cross.value) else False
        self.circle = True if keys.get(PS.circle.value) else False
        self.square = True if keys.get(PS.square.value) else False
        self.triangle = True if keys.get(PS.triangle.value) else False
        self.share = True if keys.get(PS.share.value) else False
        self.ps = True if keys.get(PS.ps.value) else False
        self.options = True if keys.get(PS.options.value) else False
        self.l1 = True if keys.get(PS.l1.value) else False
        self.l3 = True if keys.get(PS.l3.value) else False
        self.r1 = True if keys.get(PS.r1.value) else False
        self.r3 = True if keys.get(PS.r3.value) else False
        self.arrow_up = True if keys.get(PS.arrow_up.value) else False
        self.arrow_down = True if keys.get(PS.arrow_down.value) else False
        self.arrow_left = True if keys.get(PS.arrow_left.value) else False
        self.arrow_right = True if keys.get(PS.arrow_right.value) else False
        self.touch_pad = True if keys.get(PS.touch_pad.value) else False

    def get_pressed(self):
        """
        Returns a dict of all keys pressed on the controller.
        Usage:
            keys = self.get_pressed()
            if keys[<key_num>]:
                action()
        """
        self.pressed = {}
        for button_num in range(self.js.get_numbuttons()):
            self.pressed[button_num] = self.js.get_button(button_num)
        return self.pressed

    def get_left_stick(self, custom_threshold=None) -> tuple[float, float]:
        """ returns a tuple representing the position of the left joystick """
        # pygame.event.pump()
        if custom_threshold == None:
            threshold = self.threshold
        elif type(custom_threshold) in [int, float]:
            assert custom_threshold >= 0
            threshold = custom_threshold
        else:
            raise TypeError(
                f"<custom_threshold> must be either None or an Integer or a float (not '{custom_threshold}')")

        vect = pygame.Vector2((self.js.get_axis(0), self.js.get_axis(1)))
        if vect.length() < threshold:
            return (0, 0)
        else:
            return vect

    def get_left_stick_angle(self, always_positive=True, decimals=1, default_return_value=0,
                             custom_threshold=None) -> int:
        """ returns the current angle of the left joystick; 0 == down """
        assert type(always_positive) == bool
        assert type(decimals) == int

        js = self.get_right_stick(custom_threshold)
        if js == (0, 0):
            return default_return_value
        deg = round_half_up(math.degrees(math.atan2(js[0], js[1])), decimals)

        if always_positive:
            return abs(deg)
        return deg

    def get_right_stick(self, custom_threshold=None) -> tuple[float, float]:
        """ returns a tuple representing the position of the right joystick """
        if custom_threshold == None:
            threshold = self.threshold
        elif type(custom_threshold) in [int, float]:
            assert custom_threshold >= 0
            threshold = custom_threshold
        else:
            raise TypeError(
                f"<custom_threshold> must be either None or an Integer or a float (not '{custom_threshold}')")

        vect = pygame.Vector2((self.js.get_axis(2), self.js.get_axis(3)))
        if vect.length() < threshold:
            return (0, 0)
        else:
            return vect

    def get_right_stick_angle(
            self, always_positive=True, decimals=1, default_return_value=0, custom_threshold=None) -> int:
        """ returns the current angle of the right joystick; 0 == down """
        assert type(always_positive) == bool
        assert type(decimals) == int

        js = self.get_right_stick(custom_threshold)
        if js == (0, 0):
            return default_return_value
        deg = round_half_up(math.degrees(math.atan2(js[0], js[1])), decimals)

        if always_positive:
            return abs(deg)
        return deg

    def get_l2(self, decimals=1) -> float:
        """ returns a float (-1 <= x <= 1) representing how much the paddle is pressed """
        return round_half_up(self.js.get_axis(4), decimals)

    def get_r2(self, decimals=1) -> float:
        """ returns a float (-1 <= x <= 1) representing how much the paddle is pressed """
        return round_half_up(self.js.get_axis(5), decimals)



import pygame
import math

class Entity(pygame.sprite.Sprite):
    def __init__(self):
        """
        call super().__init__() right after creating self.image and self.rect

        IMPORTANT: images for animations must be a dict like:
        {
            left: img_left,
            right: img_right
        }

        self.__action__ will be priorized over self.__looping_action_   #! (is it?)

        use PygameXtras.Spritesheet for images (in self.add_action(...))
        """
        super().__init__()

        self.__entity_pos__ = [0, 0]
        self.rect = pygame.Rect(
            0, 0, self.rect.width, self.rect.height)

        self.__hitbox_data__ = (0, 0, self.rect.width, self.rect.height)
        self.__initial_hitbox_data__ = (0, 0, self.rect.width,
                                        self.rect.height)
        self.__hitbox_rect__ = pygame.Rect(
            0, 0, self.rect.width, self.rect.height)

        self.__game_map_tiles__ = None
        self.__do_tile_collision__ = False
        self.__tile_sidelength__ = None
        self.__is_platformer__ = False
        self.__is_touching_ground_precision__ = 0

        # ? idea: a dict that contains what happens during an action (eg "launch rocket at frame 7" or "die at frame 15")
        # ? the dict stores strings that look like functions and that can be called with "exec()"

        # a dict with useful data, such as "has moved", "which direction moved", "jumped", old_center, ...
        # updates every frame so all the data is accessible
        self.__data__ = {
            "old_center": (0, 0),
            "center": (0, 0),
            "last_movement": (0, 0),
            "current_movement": (0, 0),
            "has_moved": False,
            "has_moved_left": False,
            "has_moved_up": False,
            "has_moved_right": False,
            "has_moved_down": False,
            "direction": "right",
            "width": self.rect.width,
            "height": self.rect.height,
            "rect": self.rect,

            # for platformers
            "is_touching_ground": False
        }
        self.__data__["is_touching_ground_list"] = [False for i in range(
            self.__is_touching_ground_precision__)]  # [new, old, older, ...]

        # checking the init-method
        self.__init_check__ = {
            "self.add_action": False,
            "self.set_looping_action": False,
            "self.set_speed": False,
            "self.set_tile_collision": False
        }
        self.__init_check_success__ = False

        # contains the images, that are blitted next
        self.__animation_sequence__ = []

        # current 'action', change with self.__set_action__()
        self.__action__ = None
        self.__looping_action__ = None
        self.__current_action__ = None
        self.__actions__ = {}
        self.__action_run_timer__ = 0

        # additional data for movement
        self.__entity_speed__ = 0
        self.__movement_vector__ = [0, 0]
        self.__constant_movement_vector__ = [0, 0]
        self.__constant_movement_max_vector__ = [0, 0]
        self.__constant_movement_initial_vector__ = [0, 0]
        self.__constant_movement_affection__ = False
        self.__knockback_vector__ = [0, 0]
        self.__knockback_resistance__ = 1
        self.__automatic_direction_control__ = True
        self.__direction_factor__ = 1  # -1 == left; 1 == right
        self.__rotation__ = 0
        self.__rotation_point_right__ = (0, 0)
        self.__rotation_point_left__ = (0, 0)
        self.__temp_collision_rects__ = []

    def __get_distance__(self, xy1, xy2, digits_after_comma=None):
        x1, y1 = xy1[0], xy1[1]
        x2, y2 = xy2[0], xy2[1]
        if digits_after_comma == None:
            return math.sqrt(((x2-x1)**2)+((y2-y1)**2))
        else:
            return self.round_half_up(math.sqrt(((x2-x1)**2)+((y2-y1)**2)), digits_after_comma)

    def __get_angle_between_points__(self, xy1, xy2):
        vect = pygame.math.Vector2(xy2[0]-xy1[0], xy2[1]-xy1[1])
        if vect.length() != 0:
            vect.normalize()
        cos = self.__rhu__(vect.y, 2)
        deg_cos = int(math.degrees(math.acos(vect.x)))
        if cos <= 0:
            return deg_cos
        else:
            return 180 + (180-deg_cos)

    def __get_normalized_vector_between_points__(self, xy1, xy2):
        return pygame.math.Vector2(xy2[0]-xy1[0], xy2[1]-xy1[1]).normalize()

    def __get_normalized_vector_from_angle__(self, angle):
        vect = pygame.Vector2(self.__rhu__(math.cos(math.radians(
            angle)), 3), self.__rhu__(math.sin(math.radians(angle)), 3))
        return vect.normalize()

    def __rhu__(self, n, decimals=0):
        """ -> round_half_up """
        multiplier = 10 ** decimals
        return math.floor(n*multiplier + 0.5) / multiplier

    def __reset_constant_movement_vector__(self):
        self.__constant_movement_vector__ = self.__constant_movement_initial_vector__[:]

    def __move__(self):

        # x movement ##########################################################################################

        if self.__knockback_vector__[0] != 0:
            self.__movement_vector__[0] += self.__knockback_vector__[0]
            self.__knockback_vector__[0] = int(self.__knockback_vector__[0] / self.__knockback_resistance__)

        if self.__constant_movement_affection__:
            self.__movement_vector__[0] += self.__constant_movement_vector__[0]

            if (self.__movement_vector__[0] > 0 and self.__constant_movement_vector__[0] > 0) or \
                    (self.__movement_vector__[0] < 0 and self.__constant_movement_vector__[0] < 0):
                self.__constant_movement_vector__[0] += self.__constant_movement_initial_vector__[0]
            else:
                self.__constant_movement_vector__[0] = self.__constant_movement_initial_vector__[0]

            # limits the vector to its maximum value
            if self.__constant_movement_max_vector__[0] > 0:
                if self.__constant_movement_vector__[0] > self.__constant_movement_max_vector__[0]:
                    self.__constant_movement_vector__[0] = self.__constant_movement_max_vector__[0]
            elif self.__constant_movement_max_vector__[0] < 0:
                if self.__constant_movement_vector__[0] < self.__constant_movement_max_vector__[0]:
                    self.__constant_movement_vector__[0] = self.__constant_movement_max_vector__[0]

        self.rect.x += self.__movement_vector__[0]
        self.__adjust_hitbox_position__()
        if self.__do_tile_collision__:
            surrounding_tiles = self.__get_tiles_within_radius__(
                self.__game_map_tiles__, self.get_pos_factor(1/self.__tile_sidelength__))
            if len(self.__temp_collision_rects__) > 0:
                surrounding_tiles += self.__temp_collision_rects__
            collided_tiles = self.__tile_collision_test__(surrounding_tiles)
            for tile in collided_tiles:
                if self.__movement_vector__[0] > 0:
                    self.__hitbox_rect__.right = tile.left
                elif self.__movement_vector__[0] < 0:
                    self.__hitbox_rect__.left = tile.right
                self.__adjust_rect_position__()
                self.__constant_movement_vector__[0] = self.__constant_movement_initial_vector__[0]
                self.__knockback_vector__[0] = 0
        self.__movement_vector__[0] = 0

        # y movement ##########################################################################################

        if self.__knockback_vector__[1] != 0:
            self.__movement_vector__[1] += self.__knockback_vector__[1]
            self.__knockback_vector__[1] = int(self.__knockback_vector__[1] / self.__knockback_resistance__)

        if self.__constant_movement_affection__:
            self.__movement_vector__[1] += self.__constant_movement_vector__[1]

            if (self.__movement_vector__[1] > 0 and self.__constant_movement_vector__[1] > 0) or \
                    (self.__movement_vector__[1] < 0 and self.__constant_movement_vector__[1] < 0):
                self.__constant_movement_vector__[1] += self.__constant_movement_initial_vector__[1]
            else:
                self.__constant_movement_vector__[1] = self.__constant_movement_initial_vector__[1]

            # limits the vector to its maximum value
            if self.__constant_movement_max_vector__[1] > 0:
                if self.__constant_movement_vector__[1] > self.__constant_movement_max_vector__[1]:
                    self.__constant_movement_vector__[1] = self.__constant_movement_max_vector__[1]
            elif self.__constant_movement_max_vector__[1] < 0:
                if self.__constant_movement_vector__[1] < self.__constant_movement_max_vector__[1]:
                    self.__constant_movement_vector__[1] = self.__constant_movement_max_vector__[1]

        if self.__is_platformer__:
            # moving the list by one
            for i in range(self.__is_touching_ground_precision__-1, 0, -1):
                self.__data__["is_touching_ground_list"][i] = self.__data__["is_touching_ground_list"][i-1]

            self.__data__["is_touching_ground_list"][0] = False
            old_y = self.rect.y
        self.rect.y += self.__movement_vector__[1]
        self.__adjust_hitbox_position__()
        if self.__do_tile_collision__:
            surrounding_tiles = self.__get_tiles_within_radius__(
                self.__game_map_tiles__, self.get_pos_factor(1/self.__tile_sidelength__))
            if len(self.__temp_collision_rects__) > 0:
                surrounding_tiles += self.__temp_collision_rects__
            collided_tiles = self.__tile_collision_test__(surrounding_tiles)
            for tile in collided_tiles:
                if self.__movement_vector__[1] > 0:
                    self.__hitbox_rect__.bottom = tile.top
                elif self.__movement_vector__[1] < 0:
                    self.__hitbox_rect__.top = tile.bottom
                self.__adjust_rect_position__()
                self.__constant_movement_vector__[1] = self.__constant_movement_initial_vector__[1]
                self.__knockback_vector__[1] = 0

        if self.__is_platformer__:
            if old_y == self.rect.y:
                self.__data__["is_touching_ground_list"][0] = True
        self.__movement_vector__[1] = 0

        self.__temp_collision_rects__ = []

    def __update_data__(self):
        self.__data__["old_center"] = self.__data__["center"]
        self.__data__["center"] = self.get_pos()
        self.__data__["last_movement"] = (
            self.__data__["center"][0] - self.__data__["old_center"][0],
            self.__data__["center"][1] - self.__data__["old_center"][1]
        )
        self.__data__["has_moved_left"] = self.__data__["last_movement"][0] < 0
        self.__data__["has_moved_right"] = self.__data__["last_movement"][0] > 0
        self.__data__["has_moved_up"] = self.__data__["last_movement"][1] < 0
        self.__data__["has_moved_down"] = self.__data__["last_movement"][1] > 0
        self.__data__["has_moved"] = any(
            [self.__data__["has_moved_" + direction]
             for direction in "left,up,right,down".split(",")])
        if self.__automatic_direction_control__ and self.__data__["has_moved"]:
            if self.__data__["has_moved_left"]:
                self.__data__["direction"] = "left"
                self.__direction_factor__ = -1
            elif self.__data__["has_moved_right"]:
                self.__data__["direction"] = "right"
                self.__direction_factor__ = 1
        self.__data__["is_touching_ground"] = all(self.__data__["is_touching_ground_list"])

        # updating global position and position on the screen
        self.__entity_pos__ = list(self.get_pos())
        self.rect.center = self.get_pos()

    def __update_image__(self):

        self.__action_run_timer__ += 1

        static_image = False  # ! not tested (15.2.2022, 11:57)
        if self.__animation_sequence__ == []:
            if len(self.__actions__[self.__looping_action__]["frames"]) == 1:
                static_image = True
                next_image = self.__actions__[
                    self.__looping_action__]["frames"][0][
                    self.__data__["direction"]]
            else:
                self.__animation_sequence__ += self.__actions__[self.__looping_action__]["frames"]
                self.__action_run_timer__ = 0
                self.__current_action__ = self.__looping_action__

        if not static_image:
            try:
                next_image = self.__animation_sequence__[0][self.__data__["direction"]]
            except TypeError as e:
                raise Exception(
                    "At least some elements of the action '" + self.__current_action__ +
                    "' are not formatted correctly. Make sure that all elements are in the form of a dict: \n{'left': <image_left>, 'right': <image_right>}")

        if self.__rotation__ != 0:  # ! not tested (15.2.2022, 11:47)
            blue_vect = pygame.Vector2(
                self.__rotation_point_right__[0] * self.get_direction_factor(),
                self.__rotation_point_right__[1])
            saved_center = pygame.Vector2(
                self.rect.center) + pygame.Vector2(
                self.__rotation_point_right__[0] * self.get_direction_factor(),
                self.__rotation_point_right__[1])
            new_vect = blue_vect.rotate(-self.__rotation__ * self.get_direction_factor())
            next_image = pygame.transform.rotate(next_image, self.__rotation__ * self.get_direction_factor())
            self.rect = self.image.get_rect(center=saved_center - new_vect)

        # ? check hitbox stuff
        # ? check pygame.mask

        self.image = next_image

        if not static_image:
            del self.__animation_sequence__[0]

    def __execute_action_methods__(self):
        if self.__action_run_timer__ in self.__actions__[
                self.__current_action__]["methods_to_execute"].keys():
            string = "self." + self.__actions__[
                self.__current_action__]["methods_to_execute"][
                self.__action_run_timer__]
            try:
                exec(string)
            except Exception as e:
                raise Exception(
                    f"An error occurred while executing a method from the action '{self.__current_action__}':\n\n{e}")

    def __tile_collision_test__(self, tiles: list) -> list:
        """ not very precise, should only be used for movement """
        collisions = []
        for tile in tiles:
            if self.__hitbox_rect__.colliderect(tile):
                collisions.append(tile)
        return collisions

    def __get_tiles_within_radius__(self, tiles_in_2d: list[list], xy: tuple, radius: int = 2) -> list:
        """ gets awfully inefficient with increasing radius, but 2 should be sufficient for everything """
        # automatically avoids crossing game borders
        tiles = []
        for x in range(max(int(xy[0])-radius, 0),
                       min(int(xy[0])+radius, len(tiles_in_2d))):
            for y in range(max(int(xy[1])-radius, 0),
                           min(int(xy[1])+radius, len(tiles_in_2d[0]))):
                if tiles_in_2d[x][y] != 0:
                    tiles.append(tiles_in_2d[x][y])
        return tiles

    def __adjust_hitbox_position__(self):
        self.__hitbox_rect__.topleft = (
            self.rect.topleft[0] + self.__hitbox_data__[0],
            self.rect.topleft[1] + self.__hitbox_data__[1])

    def __adjust_rect_position__(self):
        self.rect.topleft = (
            self.__hitbox_rect__.topleft[0] - self.__hitbox_data__[0],
            self.__hitbox_rect__.topleft[1] - self.__hitbox_data__[1])

    def __init_check_func__(self):
        if not all(self.__init_check__.values()):
            string1 = [f"{k}: {v}" for k, v in self.__init_check__.items()]
            string2 = "\n".join(string1)
            raise Exception(f"Not all necessary init-methods called:\n\n{string2}")

    def internal_update(self):
        """ needs to be called at the end of the entities .move method \n
            methods can also be called individually """
        if self.__init_check_success__ == False:
            self.__init_check_func__()
            self.__init_check_success__ = True

        self.__move__()
        self.__update_data__()
        self.__update_image__()
        self.__execute_action_methods__()

    def add_action(self, name: str, image_dicts: list, frames_per_image: list[int],
                   methods_to_execute: dict[str: str] = None):
        """
        image_dicts: dicts of image ("left":..., "right":...)
        frames_per_image can be either a list or an int \n
        adds an action to self.__actions__ in the following form:
        {
            "<action_name>": {
                "frames": ["<frame0>", "<frame1>", "<frame2>", ...],
                "methods_to_execute": {
                    "<frame>": "self.method(*args)",
                    ...
                }
            }
        }
        """
        if type(frames_per_image) == list:
            assert len(image_dicts) == len(
                frames_per_image), "received lists of different lengths"

        # creating list
        frame_sequence = []
        if type(frames_per_image) == list:
            for image, count in zip(image_dicts, frames_per_image):
                for i in range(count):
                    frame_sequence.append(image)
        elif type(frames_per_image) == int:
            for image in image_dicts:
                for i in range(frames_per_image):
                    frame_sequence.append(image)

        # adding "methods_to_execute" functionality
        self._layer = None  # dont ask why this is here, but it crashes if it isnt
        if methods_to_execute != None:
            method_list = [func for func in dir(self) if callable(
                getattr(self, func)) and not func.startswith("__")]  # "set_action", ...
            for k, v in methods_to_execute.items():
                if type(k) != int:
                    raise ValueError(
                        f"Error in 'methods_to_execute': '{k}' is not a number")
                if v.split("(")[0] not in method_list:  # ) avoid brackets changing colors
                    # ! not sure if this works as intended
                    raise Exception(
                        f"Error in 'methods_to_execute': '{v.split('(')[0]}' is not a valid method")  # )
        else:
            methods_to_execute = {}

        self.__actions__[name] = {
            "frames": frame_sequence,
            "methods_to_execute": methods_to_execute
        }
        self.__init_check__["self.add_action"] = True

    def set_looping_action(self, action: str, cancel_other_action: bool = True) -> None:
        assert action in self.__actions__.keys(), f"unknown action: '{action}'"
        if cancel_other_action and self.__looping_action__ != action:
            self.__animation_sequence__ = []
            self.__animation_sequence__ += self.__actions__[action]["frames"]
            self.__action_run_timer__ = 0
            self.__current_action__ = action
        self.__looping_action__ = action
        self.__init_check__["self.set_looping_action"] = True

    def set_action(self, action: str, cancel_other_action: bool = True) -> None:
        assert action in self.__actions__.keys(), f"unknown action: '{action}'"
        if cancel_other_action and self.__current_action__ != action:
            self.__animation_sequence__ = []
            self.__animation_sequence__ += self.__actions__[action]["frames"]
            self.__action_run_timer__ = 0
            self.__current_action__ = action
        self.__action__ = action

    def get_data(self, name: str):
        """ access useful data like "has_moved", "last_movement", etc. """
        assert name in self.__data__.keys(), f"no entry in __data__ for '{name}'"
        return self.__data__[name]

    def set_speed(self, speed: float):
        """ needs to be called in the init-method """
        self.__entity_speed__ = speed
        self.__init_check__["self.set_speed"] = True

    def get_speed(self):
        return self.__entity_speed__

    def move_right(self, custom_value=None):
        if custom_value == 0:
            return
        elif custom_value == None:
            self.__movement_vector__[0] += self.__entity_speed__
        elif type(custom_value) in [int, float]:
            self.__movement_vector__[0] += custom_value
        else:
            raise TypeError(f"Incorrect input type for 'custom_value': {custom_value}")

    def move_left(self, custom_value=None):
        if custom_value == 0:
            return
        elif custom_value == None:
            self.__movement_vector__[0] -= self.__entity_speed__
        elif type(custom_value) in [int, float]:
            self.__movement_vector__[0] -= custom_value
        else:
            raise TypeError(f"Incorrect input type for 'custom_value': {custom_value}")

    def move_up(self, custom_value=None):
        if custom_value == 0:
            return
        elif custom_value == None:
            self.__movement_vector__[1] -= self.__entity_speed__
        elif type(custom_value) in [int, float]:
            self.__movement_vector__[1] -= custom_value
        else:
            raise TypeError(f"Incorrect input type for 'custom_value': {custom_value}")

    def move_down(self, custom_value=None):
        if custom_value == 0:
            return
        elif custom_value == None:
            self.__movement_vector__[1] += self.__entity_speed__
        elif type(custom_value) in [int, float]:
            self.__movement_vector__[1] += custom_value
        else:
            raise TypeError(f"Incorrect input type for 'custom_value': {custom_value}")

    def move_horizontal(self, custom_value=None):
        if custom_value == 0:
            return
        elif custom_value == None:
            self.__movement_vector__[0] += self.__entity_speed__
        elif type(custom_value) in [int, float]:
            self.__movement_vector__[0] += custom_value
        else:
            raise TypeError(f"Incorrect input type for 'custom_value': {custom_value}")

    def move_vertical(self, custom_value=None):
        if custom_value == 0:
            return
        elif custom_value == None:
            self.__movement_vector__[1] += self.__entity_speed__
        elif type(custom_value) in [int, float]:
            self.__movement_vector__[1] += custom_value
        else:
            raise TypeError(f"Incorrect input type for 'custom_value': {custom_value}")

    def get_pos(self, reference_point="center"):
        """ uses pixel """
        assert reference_point in ["topleft", "top", "topright",
                                   "left", "center", "centerx", "centery", "right",
                                   "bottomleft", "bottom", "bottomright"]

        if reference_point == "topleft":
            return self.rect.topleft
        elif reference_point == "top":
            return self.rect.top
        elif reference_point == "topright":
            return self.rect.topright
        elif reference_point == "left":
            return self.rect.left
        elif reference_point == "center":
            return self.rect.center
        elif reference_point == "centerx":
            return self.rect.centerx
        elif reference_point == "centery":
            return self.rect.centery
        elif reference_point == "right":
            return self.rect.right
        elif reference_point == "bottomleft":
            return self.rect.bottomleft
        elif reference_point == "bottom":
            return self.rect.bottom
        elif reference_point == "bottomright":
            return self.rect.bottomright

    def get_pos_factor(self, factor, reference_point="center", decimals=0):
        """ uses the coordinate system """
        assert reference_point in ["topleft", "top", "topright",
                                   "left", "center", "centerx", "centery", "right",
                                   "bottomleft", "bottom", "bottomright"]

        if reference_point == "topleft":
            return self.__rhu__(
                self.rect.topleft[0] * factor, decimals), self.__rhu__(
                self.rect.topleft[1],
                decimals)
        elif reference_point == "top":
            return self.__rhu__(self.rect.top * factor, decimals)
        elif reference_point == "topright":
            return self.__rhu__(
                self.rect.topright[0] * factor, decimals), self.__rhu__(
                self.rect.topright[1] * factor, decimals)
        elif reference_point == "left":
            return self.__rhu__(self.rect.left * factor, decimals)
        elif reference_point == "center":
            return self.__rhu__(
                self.rect.center[0] * factor, decimals), self.__rhu__(
                self.rect.center[1] * factor, decimals)
        elif reference_point == "centerx":
            return self.__rhu__(self.rect.centerx * factor, decimals)
        elif reference_point == "centery":
            return self.__rhu__(self.rect.centery * factor, decimals)
        elif reference_point == "right":
            return self.__rhu__(self.rect.right * factor, decimals)
        elif reference_point == "bottomleft":
            return self.__rhu__(
                self.rect.bottomleft[0] * factor, decimals), self.__rhu__(
                self.rect.bottomleft[1] * factor, decimals)
        elif reference_point == "bottom":
            return self.__rhu__(self.rect.bottom * factor, decimals)
        elif reference_point == "bottomright":
            return self.__rhu__(
                self.rect.bottomright[0] * factor, decimals), self.__rhu__(
                self.rect.bottomright[1] * factor, decimals)

    def set_pos(self, xy, reference_point="center"):
        """ uses pixel \n 'xy' can be either a tuple or an integer, depending on 'reference_point' """
        assert type(xy) in [tuple, list, int, pygame.Vector2]
        assert type(reference_point) == str
        if reference_point in ["topleft", "topright", "center", "bottomleft", "bottomright"]:
            assert type(xy) in [
                tuple, list, int, pygame.Vector2], f"invalid combination of <xy> ({xy}) and <reference_point> ({reference_point})"
            assert len(xy) == 2, f"invalid length of tuple <xy> ({xy})"
        elif reference_point in ["top", "left", "centerx", "centery", "right", "bottom"]:
            assert type(
                xy) == int, f"invalid combination of <xy> ({xy}) and <reference_point> ({reference_point})"

        if reference_point == "topleft":
            self.rect.topleft = xy
        elif reference_point == "top":
            self.rect.top = xy
        elif reference_point == "topright":
            self.rect.topright = xy
        elif reference_point == "left":
            self.rect.left = xy
        elif reference_point == "center":
            self.rect.center = xy
        elif reference_point == "centerx":
            self.rect.centerx = xy
        elif reference_point == "centery":
            self.rect.centery = xy
        elif reference_point == "right":
            self.rect.right = xy
        elif reference_point == "bottomleft":
            self.rect.bottomleft = xy
        elif reference_point == "bottom":
            self.rect.bottom = xy
        elif reference_point == "bottomright":
            self.rect.bottomright = xy

        self.__entity_pos__ = list(self.rect.center)

    def set_pos_factor(self, xy, factor: int, reference_point="center"):
        """ uses the coordinate system if <factor> == self.__tile_sidelength__ \n <xy> can be either a tuple or an integer, depending on <reference_point> """
        assert type(xy) in [tuple, list, int]
        assert type(reference_point) == str
        # <factor> could (and should!) be automated with the side length of the current levels tiles
        if reference_point in ["topleft", "topright", "center", "bottomleft", "bottomright"]:
            assert type(xy) in [
                list, tuple], f"invalid combination of <xy> ({xy}) and <reference_point> ({reference_point})"
            assert len(xy) == 2, f"invalid length of tuple <xy> ({xy})"
        elif reference_point in ["top", "left", "centerx", "centery", "right", "bottom"]:
            assert type(
                xy) == int, f"invalid combination of <xy> ({xy}) and <reference_point> ({reference_point})"

        assert type(factor) == int, f"invalid argument for <factor> ({factor}), integer required"

        if reference_point == "topleft":
            self.rect.topleft = (xy[0] * factor, xy[1] * factor)
        elif reference_point == "top":
            self.rect.top = xy * factor
        elif reference_point == "topright":
            self.rect.topright = (xy[0] * factor, xy[1] * factor)
        elif reference_point == "left":
            self.rect.left = xy * factor
        elif reference_point == "center":
            self.rect.center = (xy[0] * factor, xy[1] * factor)
        elif reference_point == "centerx":
            self.rect.centerx = xy * factor
        elif reference_point == "centery":
            self.rect.centery = xy * factor
        elif reference_point == "right":
            self.rect.right = xy * factor
        elif reference_point == "bottomleft":
            self.rect.bottomleft = (xy[0] * factor, xy[1] * factor)
        elif reference_point == "bottom":
            self.rect.bottom = xy * factor
        elif reference_point == "bottomright":
            self.rect.bottomright = (xy[0] * factor, xy[1] * factor)

        self.__entity_pos__ = list(self.rect.center)

    def move_to(self, xy, reference_point="center"):
        """ same as self.set_pos() """
        self.set_pos(xy, reference_point)

    def set_game_map_tiles(self, tiles: list[list], do_tile_collision: bool = True):
        """ <tiles> must be a 2d list of pygame.Rect objects """
        assert type(tiles) == list, "failed to load map tiles"
        assert type(tiles[0]) == list, "failed to load map tiles"
        assert type(do_tile_collision) == bool
        self.__game_map_tiles__ = tiles
        self.__do_tile_collision__ = do_tile_collision

    def set_tile_collision(self, boolean: bool, tile_sidelength: int = None):
        assert type(boolean) == bool
        self.__do_tile_collision__ = boolean
        if boolean == True:
            assert tile_sidelength != None, "if tile_collision should be done, tile_sidelength needs to be set"
            assert type(tile_sidelength) == int
            assert tile_sidelength >= 8, "tile_sidelength must be greater or equal to 8"
            self.set_tile_sidelength(tile_sidelength)
        self.__init_check__["self.set_tile_collision"] = True

    def get_tile_sidelength(self):
        return self.__tile_sidelength__

    def set_constant_movement(self, vector2: tuple, max_vector2: tuple,
                              set_constant_movement_affection: bool = True):
        assert type(vector2) in [list, tuple, pygame.Vector2], f"invalid vector ({vector2})"
        assert len(vector2) == 2, f"invalid vector ({vector2})"
        assert type(max_vector2) in [list, tuple, pygame.Vector2], f"invalid vector ({max_vector2})"
        assert len(max_vector2) == 2, f"invalid vector ({max_vector2})"
        assert type(set_constant_movement_affection) == bool

        self.__constant_movement_vector__ = list(vector2)
        self.__constant_movement_max_vector__ = list(max_vector2)
        self.__constant_movement_initial_vector__ = list(vector2)
        self.set_constant_movement_affection(set_constant_movement_affection)

    def get_constant_movement(self):
        return self.__constant_movement_vector__

    def set_constant_movement_affection(self, boolean):
        assert type(boolean) == bool
        self.__constant_movement_affection__ = boolean

    def get_constant_movement_affection(self):
        return self.__constant_movement_affection__

    def set_custom_hitbox(self, rect: tuple[int, int, int, int]):
        """ sets a custom hitbox for tile collision """
        assert type(rect) in [list, tuple]
        assert len(rect) == 4

        self.__hitbox_data__ = rect
        self.__hitbox_rect__ = pygame.Rect(*rect)
        self.__adjust_hitbox_position__()

    def get_hitbox(self):
        return self.__hitbox_rect__

    def reset_hitbox(self):
        """ resets the hitbox to default (the entities 'rect') """
        self.__hitbox_data__ = self.__initial_hitbox_data__
        self.__hitbox_rect__ = pygame.Rect(*self.__hitbox_data__)
        self.__adjust_hitbox_position__()

    def set_knockback_resistance(self, knockback_resistance: float):
        """ sets the number the knockback vector will be divided by each frame, must be between 1 and 1.2 """
        assert 1 < knockback_resistance <= 1.2, "knockback_resistance must be between 1 and 1.2"
        self.__knockback_resistance__ = knockback_resistance

    def get_knockback_resistance(self):
        """ returns the number the knockback vector gets divided by """
        return self.__knockback_resistance__

    def set_knockback(self, vector2: tuple, reset_constant_movement: bool = True):
        """ applies knockback to the entity """
        assert type(vector2) in [list, tuple, pygame.Vector2], f"invalid vector ({vector2})"
        assert len(vector2) == 2, f"invalid vector ({vector2})"
        assert self.__knockback_resistance__ != 1, "knockback_resistance must be between 1 and 1.2"

        self.__knockback_vector__ = list(vector2)
        if reset_constant_movement:
            self.__reset_constant_movement_vector__()

    def add_knockback(self, vector2: tuple):
        """ applies additional knockback on top of the current one to the entity """
        assert type(vector2) in [list, tuple, pygame.Vector2], f"invalid vector ({vector2})"
        assert len(vector2) == 2, f"invalid vector ({vector2})"
        assert self.__knockback_resistance__ != 1, "knockback_resistance must be between 1 and 1.2 "

        self.__knockback_vector__ = [
            self.__knockback_vector__[0] + vector2[0],
            self.__knockback_vector__[1] + vector2[1]]

    def move_at_angle(self, angle, custom_speed=None):
        """ 0 degrees = right | then clockwise """
        temp_vect = self.__get_normalized_vector_from_angle__(angle)
        if custom_speed != None:
            vect = temp_vect * custom_speed
        else:
            vect = temp_vect * self.__entity_speed__

        self.__movement_vector__[0] += vect[0]
        self.__movement_vector__[1] += vect[1]

    def set_tile_sidelength(self, tile_sidelength: int):
        assert type(tile_sidelength) == int
        self.__tile_sidelength__ = tile_sidelength

    def set_platformer_status(self, is_platformer: bool, set_touching_ground_precision: int = 3):
        """ decide whether the game is a platformer or not (enables some platformer features) """
        assert type(is_platformer) == bool
        self.__is_platformer__ = is_platformer
        self.set_touching_ground_precision(set_touching_ground_precision)

    def set_touching_ground_precision(self, precision: int):
        """ higher precision means that the ground will be detected with more certainty, 
        but it takes longer to realize that the entity is not touching the ground anymore...
        this might result in the player being able to jump again mid-air \n
        note: \n
        automatically calls self.set_platformer_status
        1 <= precision <= 6 | recommended value: 3 """
        assert type(precision) == int
        assert 1 <= precision <= 6
        self.__is_platformer__ = True
        self.__is_touching_ground_precision__ = precision
        self.__data__["is_touching_ground_list"] = [False for i in range(
            self.__is_touching_ground_precision__)]  # [new, old, older, ...]

    def set_automatic_direction_control(self, boolean: bool):
        """ automatically sets the direction of the entity based on its movement """
        assert type(boolean) == bool
        self.__automatic_direction_control__ = boolean

    def set_direction(self, direction):
        """ <direction> can be either "left" and "right" or -1 and 1 """
        assert direction in ["left", "right", -1, 1]
        if direction in ["left", -1]:
            self.__data__["direction"] = "left"
            self.__direction_factor__ = -1
        elif direction in ["right", 1]:
            self.__data__["direction"] = "right"
            self.__direction_factor__ = 1

    def get_direction(self):
        return self.get_data("direction")

    def get_direction_factor(self):
        return self.__direction_factor__

    def set_rotation_point_vector(self, xy: tuple):
        """ sets the point the entity rotates around; vector starts at the center of the entity; \n
            default = center; IMPORTANT: configure for entity facing right """
        assert type(xy) in [tuple, list, pygame.Vector2]
        self.__rotation_point_right__ = tuple(xy)
        self.__rotation_point_left__ = (xy[0]*-1, xy[1])

    def get_rotation_point_right(self) -> tuple:
        """ returns the point the entity rotates around when it is facing right """
        return self.__rotation_point_right__

    def get_rotation_point_left(self) -> tuple:
        """ returns the point the entity rotates around when it is facing left """
        return self.__rotation_point_left__

    def set_rotation(self, angle):
        """ sets the entities rotation to the given angle """
        assert type(angle) in [int, float]
        self.__rotation__ = int(angle) % 360

    def get_rotation(self):
        """ returns the current rotation """
        return self.__rotation__

    def do_platformer_jump(self, factor: float):
        self.__constant_movement_vector__[1] = factor

    def add_temp_collision_rects(self, rects: list):
        """ adds a list of rectangles to the collision tiles list for ONE iteration """
        assert type(rects) in [tuple, list]
        for item in rects:
            assert type(item) == pygame.Rect
        self.__temp_collision_rects__ = rects



import pygame
from .Button import Button

class ScrollableButtonList:
    def __init__(
            self, surface, target_rect: tuple[int, int, int, int], scrolling_speed: int, backgroundcolor: tuple = (0,0,0)):
        """
        target_rect specifies the area in which buttons will be visible

        to update the buttons, use: 
            for b in self.get_buttons():
                if b.update(event_list, offset=self.get_offset()):
                    ...

        also, a general update every loop is needed:
            self.update(event_list)

        it is also possible to change the buttons:
            self.clear_buttons()
            self.add_buttons(list_of_button_names)

        IMPORTANT:
        this widget has to be drawn DIRECTLY to the screen, otherwise the clicking will be messed up.
        to check the active area, use self.get_rect() and draw it to the screen
        ALSO IMPORTANT:
        the buttons might not show up if the self.update method is not constantly called. to
        fix this, just call the method once with an emtpy list after initializing the class. (self.update([]))
        """

        # only vertical scroll
        # TODO: assertions
        assert type(surface) == pygame.Surface, f"invalid argument for 'surface': {surface}"
        self.__main_surface__ = surface
        assert type(target_rect) in [tuple, list, pygame.Rect]
        if type(target_rect) in [tuple, list]:
            assert len(target_rect) == 4, f"invalid argument for 'target_rect': {target_rect}"
        self.__target_rect__ = pygame.Rect(target_rect)
        self.__surface__ = pygame.Surface((self.__target_rect__.width, 0))

        assert type(scrolling_speed) == int, f"invalid argument for 'scrolling_speed': {scrolling_speed}"
        assert scrolling_speed > 0, f"invalid argument for 'scrolling_speed': {scrolling_speed}"
        self.__scrolling_speed__ = scrolling_speed

        assert type(backgroundcolor) in [tuple, list], f"invalid argument for 'backgroundcolor': {backgroundcolor}"
        assert len(backgroundcolor) == 3, f"invalid argument for 'backgroundcolor': {backgroundcolor}"
        self.__backgroundcolor__ = backgroundcolor        

        self.__scroll__ = 0
        self.__max_scroll__ = self.__surface__.get_height() - self.__target_rect__.height #! has to be reconfigured if anything changes

        self.__button_names__: list[str] = []
        self.__old_button_names__: list[str] = []
        self.__buttons__: list[Button] = []

        self.__int_surf__ = pygame.Surface((10, 10)) # internal surface

    def set_button_style(self, size: int, **kwargs):
        """ dont forget about the backgroundcolor and the height """
        try:
            Button(self.__int_surf__, "Test", size=size, xy=(0,0), anchor="topleft", **kwargs)
            self.__size__ = size
            kwargs["bR"] = 1 # important !
            kwargs["fw"] = self.__target_rect__.width
            kwargs["aA"] = self.__target_rect__
            self.__style__ = kwargs
        except Exception as e:
            raise e

    def add_button(self, button_name: str):
        self.__button_names__.append(str(button_name))

    def add_buttons(self, button_names: list[str]):
        for button in button_names:
            self.add_button(button)

    def clear_buttons(self):
        """ removes all buttons """
        self.__button_names__ = []
    
    def reset_scroll(self):
        """ resets the scroll value """
        self.__scroll__ = 0

    def set_buttons(self, button_names: list[str]):
        self.__button_names__ = []
        self.add_buttons(button_names)

    def update_surface(self):
        """ updates the surface if something has changed """

        if self.__old_button_names__ != self.__button_names__:
            self.__old_button_names__ = self.__button_names__[:]
            self.__buttons__ = []
            borderwidth = 0
            for count, name in enumerate(self.__button_names__):
                if count == 0:
                    self.__buttons__.append(
                        Button(self.__int_surf__, name, self.__size__, (0,0), "topleft", **self.__style__)
                    )
                    borderwidth = self.__buttons__[0].borderwidth
                else:
                    self.__buttons__.append(
                        Button(self.__int_surf__, name, self.__size__, (self.__buttons__[count-1].left, self.__buttons__[count-1].bottom-borderwidth), "topleft", **self.__style__)
                    )

            # determining the surface height
            if len(self.__buttons__) > 0:
                lowest = self.__buttons__[-1].bottom
            else:
                lowest = 0
            self.__surface__ = pygame.Surface((self.__target_rect__.width, lowest))
            
            self.__max_scroll__ = self.__surface__.get_height() - self.__target_rect__.height
            if self.__max_scroll__ > 0 and self.__scroll__ > self.__max_scroll__:
                self.__scroll__ = self.__max_scroll__
            elif self.__max_scroll__ <= 0:
                self.__scroll__ = 0

    def update_scroll(self, event_list):
        """ updates the scroll value """
        
        if self.__max_scroll__ > 0:
            for event in event_list:
                if event.type == pygame.MOUSEWHEEL:
                    self.__scroll__ = max(
                        min(self.__max_scroll__, self.__scroll__ + event.y*-1*self.__scrolling_speed__), 0)

    def update(self, event_list):
        """ updates surface and scroll """
        self.update_surface()
        self.update_scroll(event_list)

    def draw(self):
        self.__surface__.fill(self.__backgroundcolor__)
        for button in self.__buttons__:
            button.draw_to(self.__surface__)
        tr = self.__target_rect__
        pygame.draw.rect(self.__main_surface__, self.__backgroundcolor__, tr)
        if self.__surface__.get_height() > tr.height:
            self.__main_surface__.blit(self.__surface__.subsurface(
                (0, self.__scroll__, tr[2], tr[3])), self.__target_rect__.topleft)
        else:
            self.__main_surface__.blit(self.__surface__, self.__target_rect__.topleft)

    def get_buttons(self):
        return self.__buttons__

    def get_offset(self):
        return self.__target_rect__[0], self.__target_rect__[1] - self.__get_scroll__()

    def __get_scroll__(self):
        return self.__scroll__

    def get_rect(self):
        return self.__target_rect__


import pygame

class Bar:
    def __init__(self, surface, width_height: tuple, xy: tuple, anchor="center", **kwargs):
        """
        Creates a bar.

        Instructions:
            - To create a bar, create an instance of this class before
            the mainloop of the game.
            - To make the bar appear, call the method 'self.draw()' in
            every loop of the game.
            - To update the fill percentage of the bar, call the
            self.update() method before drawing to the screen.

        Example (simplified):
            bar = Bar(screen, (100, 20), (100, 100))
            health = 20
            max_health = 100
            while True:
                bar.update(health, max_health)
                bar.draw()

        Arguments:
            surface
                the surface the object will be drawn onto
                Type: pygame.Surface, pygame.display.set_mode
            width_height
                specifies the width / height of the bar
                Type: tuple, list
            xy
                refers to the position of the anchor
                Type: tuple, list
            anchor
                the anchor of the object:   topleft,    midtop,    topright,
                                            midleft,    center,    midright,
                                            bottomleft, midbottom, bottomright
                Type: str
            backgroundcolor
                color of the background in rgb
                Type: None, tuple, list
            fillcolor
                color of the filling
                Type: tuple, list
            bordercolor
                color of the border
                Type: tuple, list
            borderwidth
                width of the border
                Type: int
            borderradius
                used to round the corners of the background_rect; use int for all corners or other types for individual corners
                Type: int, tuple, list
            info
                can store any kind of data; usage of dict recommended
                Type: Any
            fill_start
                from which side of the bar to start filling: left (default), right, top, bottom
                Type: str

        All custom arguments can also be used in their short form (eg. "aa" instead of "antialias").
        To see what all the short forms look like, inspect the self.ABBREVIATIONS attribute.
        """

        ABBREVIATIONS = {
            "backgroundcolor": "bgc",
            "fillcolor": "fc",
            "bordercolor": "bc",
            "borderwidth": "bw",
            "borderradius": "br",
            "info": "info",
            "fill_start": "fs"
        }

        self.__surface = surface
        self.__width_height = width_height
        self.__xy = xy
        self.__anchor = anchor
    
        # backgroundcolor #
        self.__backgroundcolor = kwargs.get('backgroundcolor', (0, 0, 0))
        if self.__backgroundcolor == (0, 0, 0):
            self.__backgroundcolor = kwargs.get(ABBREVIATIONS['backgroundcolor'], (0, 0, 0))
        # assertion #
        if self.__backgroundcolor != (0, 0, 0):
            assert isinstance(self.__backgroundcolor, (tuple, list)), f"invalid argument for 'backgroundcolor': {self.__backgroundcolor}"
            assert len(self.__backgroundcolor) == 3, f"invalid argument for 'backgroundcolor': {self.__backgroundcolor}"
        
        # fillcolor #
        self.__fillcolor = kwargs.get('fillcolor', (255, 255, 255))
        if self.__fillcolor == (255, 255, 255):
            self.__fillcolor = kwargs.get(ABBREVIATIONS['fillcolor'], (255, 255, 255))
        # assertion #
        if self.__fillcolor != (255, 255, 255):
            assert isinstance(self.__fillcolor, (tuple, list)), f"invalid argument for 'fillcolor': {self.__fillcolor}"
            assert len(self.__fillcolor) == 3, f"invalid argument for 'fillcolor': {self.__fillcolor}"
        
        # bordercolor #
        self.__bordercolor = kwargs.get('bordercolor', (0, 0, 0))
        if self.__bordercolor == (0, 0, 0):
            self.__bordercolor = kwargs.get(ABBREVIATIONS['bordercolor'], (0, 0, 0))
        # assertion #
        if self.__bordercolor != (0, 0, 0):
            assert isinstance(self.__bordercolor, (tuple, list)), f"invalid argument for 'bordercolor': {self.__bordercolor}"
            assert len(self.__bordercolor) == 3, f"invalid argument for 'bordercolor': {self.__bordercolor}"
        
        # borderwidth #
        self.__borderwidth = kwargs.get('borderwidth', 0)
        if self.__borderwidth == 0:
            self.__borderwidth = kwargs.get(ABBREVIATIONS['borderwidth'], 0)
        # assertion #
        if self.__borderwidth != 0:
            assert isinstance(self.__borderwidth, (int)), f"invalid argument for 'borderwidth': {self.__borderwidth}"
        
        # borderradius #
        self.__borderradius = kwargs.get('borderradius', 1)
        if self.__borderradius == 1:
            self.__borderradius = kwargs.get(ABBREVIATIONS['borderradius'], 1)
        # assertion #
        if self.__borderradius != 1:
            assert isinstance(self.__borderradius, (int)), f"invalid argument for 'borderradius': {self.__borderradius}"
        
        # info #
        self.__info = kwargs.get('info', None)
        if self.__info == None:
            self.__info = kwargs.get(ABBREVIATIONS['info'], None)
        # assertion #
        if self.__info != None:
            assert isinstance(self.__info, (None)), f"invalid argument for 'info': {self.__info}"
        
        # fill_start #
        self.__fill_start = kwargs.get('fill_start', 'left')
        if self.__fill_start == 'left':
            self.__fill_start = kwargs.get(ABBREVIATIONS['fill_start'], 'left')
        # assertion #
        if self.__fill_start != 'left':
            assert self.__fill_start in ('top', 'right', 'bottom', 'left'), f"invalid argument for 'fill_start': {self.__fill_start}"


        assert self.__width_height[0] - 2 * self.__borderwidth > 0, "widget is too small"
        assert self.__width_height[1] - 2 * self.__borderwidth > 0, "widget is too small"

        self.__create()

    def __create(self):
        # background
        self.__r_background = pygame.Rect(0,0, self.__width_height[0], self.__width_height[1])
        # foreground (for calculations)
        self.__r_foreground = pygame.Rect(0,0, self.__width_height[0] - 2 * self.__borderwidth, self.__width_height[1] - 2 * self.__borderwidth)
        # filling
        self.__r_filling = pygame.Rect(0,0, self.__r_foreground.width, self.__r_foreground.height)

        self.update_pos(self.__xy, self.__anchor)

    def update_pos(self, xy, anchor=None):
        """
        Changes the widgets position. Call this
        method before drawing to the screen.
        """
        assert isinstance(xy, (tuple, list)), f"invalid argument for 'xy': {xy}"
        assert len(xy) == 2, f"invalid argument for 'xy': {xy}"
        self.__xy = xy
        assert anchor in ["topleft", "topright", "bottomleft", "bottomright",
                                   "center", "midtop", "midright", "midbottom", "midleft", None]
        if anchor is not None:
            self.__anchor = anchor

        self.__r_background.__setattr__(self.__anchor, self.__xy)
        self.__r_foreground.center = self.__r_background.center
        self.__update_filling_pos()

    def __update_filling_pos(self):
        self.__r_filling.__setattr__("mid" + self.__fill_start, self.__r_foreground.__getattribute__("mid" + self.__fill_start))

    def update(self, value, max_value):
        fill = min(1, max(0, value / max_value))
        if self.__fill_start in ["left", "right"]:
            self.__r_filling.width = self.__r_foreground.width * fill
        elif self.__fill_start in ["top", "bottom"]:
            self.__r_filling.height = self.__r_foreground.height * fill
        self.__update_filling_pos()

    def draw(self):
        pygame.draw.rect(self.__surface, self.__backgroundcolor, self.__r_background, 0, self.__borderradius)
        pygame.draw.rect(self.__surface, self.__fillcolor, self.__r_filling, 0, self.__borderradius)
        if self.__borderwidth > 0:
            pygame.draw.rect(self.__surface, self.__bordercolor, self.__r_background, self.__borderwidth, self.__borderradius)

    def update_colors(self, backgroundcolor=None, fillcolor=None, bordercolor=None):
        if backgroundcolor is not None:
            assert isinstance(self.__backgroundcolor, (tuple, list)), f"invalid argument for 'backgroundcolor': {self.__backgroundcolor}"
            assert len(self.__backgroundcolor) == 3, f"invalid argument for 'backgroundcolor': {self.__backgroundcolor}"
            self.__backgroundcolor = backgroundcolor
        if fillcolor is not None:
            assert isinstance(self.__fillcolor, (tuple, list)), f"invalid argument for 'fillcolor': {self.__fillcolor}"
            assert len(self.__fillcolor) == 3, f"invalid argument for 'fillcolor': {self.__fillcolor}"
            self.__fillcolor = fillcolor
        if bordercolor is not None:
            assert isinstance(self.__bordercolor, (tuple, list)), f"invalid argument for 'bordercolor': {self.__bordercolor}"
            assert len(self.__bordercolor) == 3, f"invalid argument for 'bordercolor': {self.__bordercolor}"
            self.__bordercolor = bordercolor
        


from src.classes.Label import Label

import pygame
import sys

c1 = {
    'black': (0,0,0),
    'white': (255,255,255),
    'cyan': (0,255,255),
    'magenta': (255,0,255),
    'yellow': (255,255,0),
    'red': (255,0,0),
    'green': (0,255,0),
    'blue': (0,0,255)
}

pygame.init()
screen = pygame.display.set_mode((500,500))
fpsclock = pygame.time.Clock()
fps = 60

def circle(xy, color, radius=3, width=0):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.circle(screen, color, xy, radius, width)

def line(xy1, xy2, color, width=1):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.line(screen, color, xy1, xy2, width)

def rect(r, color, width=0):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.rect(screen, color, r, width)

center = (250, 250)

pos1 = (50, 50)
pos2 = (450, 50)
pos3 = center
pos4 = (50, 450)
pos5 = (450, 450)

# l1 = Label(screen, "Test Label 1", 25, pos1, "topleft", bw=1, fd=(120, 60), p=5)
# l2 = Label(screen, "Test Label 2", 25, pos2, "topright", bw=1, fd=(120, 60), p=5)
l3 = Label(screen, "Test Label 3", 25, pos3, "center", bw=1, fd=(300, 160), p=5, bgc=(200,80,80))
# l4 = Label(screen, "Test Label 4", 25, pos4, "bottomleft", bw=1, fd=(120, 60), p=5)
# l5 = Label(screen, "Test Label 5", 25, pos5, "bottomright", bw=1, fd=(120, 60), p=5)

while True:
    events = pygame.event.get()
    for event in events:
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            pygame.quit()
            sys.exit()

    screen.fill((100,100,200))
    # for l in [l1,l2,l3,l4,l5]:
        # l.draw()
    l3.draw()
    r = pygame.Rect(0,0,300,160)
    r.center = center
    rect(r, "red", 1)

    pygame.display.flip()
    fpsclock.tick(fps)


import unittest

from ..src.classes.Collisions import Collisions


class TestCollisions(unittest.TestCase):
    def test_circle_circle(self):
        # Positive test: circles touching
        self.assertTrue(Collisions.circle_circle((0, 0), 5, (10, 0), 5))
        # Negative test: circles not touching
        self.assertFalse(Collisions.circle_circle((0, 0), 5, (11, 0), 5))

    def test_rect_rect(self):
        # Positive test: rectangles touching
        self.assertTrue(Collisions.rect_rect((0, 0, 10, 10), (10, 0, 10, 10)))
        # Negative test: rectangles not touching
        self.assertFalse(Collisions.rect_rect((0, 0, 10, 10), (11, 0, 10, 10)))

    def test_circle_rect(self):
        # Positive test: circle touching rectangle
        self.assertTrue(Collisions.circle_rect((5, 5), 5, (0, 0, 10, 10)))
        # Negative test: circle not touching rectangle
        self.assertFalse(Collisions.circle_rect((20, 20), 5, (0, 0, 10, 10)))

    def test_line_circle(self):
        # Positive test: line touching circle
        self.assertTrue(Collisions.line_circle((0, 0), (10, 0), (5, 5), 5))
        # Negative test: line not touching circle
        self.assertFalse(Collisions.line_circle((0, 0), (10, 0), (20, 20), 5))

    def test_line_rect(self):
        # Positive test: line intersecting rectangle
        self.assertTrue(Collisions.line_rect((0, 5), (10, 5), (2, 2, 6, 6)))
        # Negative test: line not intersecting rectangle
        self.assertFalse(Collisions.line_rect((0, 0), (1, 1), (2, 2, 6, 6)))

    def test_polygon_circle(self):
        # Positive test: polygon touching circle
        self.assertTrue(
            Collisions.polygon_circle(((0, 0), (10, 0), (5, 10)), (5, 5), 5)
        )
        # Negative test: polygon not touching circle
        self.assertFalse(
            Collisions.polygon_circle(((0, 0), (10, 0), (5, 10)), (20, 20), 5)
        )

    def test_polygon_rect(self):
        # Positive test: polygon touching rectangle
        self.assertTrue(
            Collisions.polygon_rect(((0, 0), (10, 0), (5, 10)), (2, 2, 6, 6))
        )
        # Negative test: polygon not touching rectangle
        self.assertFalse(
            Collisions.polygon_rect(((0, 0), (10, 0), (5, 10)), (20, 20, 6, 6))
        )

    def test_polygon_line(self):
        # Positive test: polygon edge intersecting line
        self.assertTrue(
            Collisions.polygon_line(((0, 0), (10, 0), (5, 10)), (5, 5), (15, 5))
        )
        # Negative test: polygon edge not intersecting line
        self.assertFalse(
            Collisions.polygon_line(((0, 0), (10, 0), (5, 10)), (20, 20), (25, 25))
        )

    def test_polygon_polygon(self):
        # Positive test: polygons touching
        self.assertTrue(
            Collisions.polygon_polygon(
                ((0, 0), (10, 0), (5, 10)), ((5, 5), (15, 5), (10, 15))
            )
        )
        # Negative test: polygons not touching
        self.assertFalse(
            Collisions.polygon_polygon(
                ((0, 0), (10, 0), (5, 10)), ((20, 20), (30, 20), (25, 30))
            )
        )

    def test_line_line(self):
        # Positive test: lines intersecting
        self.assertTrue(Collisions.line_line((0, 0), (10, 10), (0, 10), (10, 0)))
        # Negative test: lines not intersecting
        self.assertFalse(Collisions.line_line((0, 0), (10, 0), (0, 10), (10, 10)))


if __name__ == "__main__":
    unittest.main()



import pygame
pygame.init()

from .psvg_util import *


class PSVG:

    __color = (255, 255, 255)
    __size = 50
    __linewidth = 5

    __up = up(__size)
    __right = right(__size)
    __down = down(__size)
    __left = left(__size)

    __triangle = triangle(__size)
    # __circle = circle(__size)
    # __cross = cross(__size)
    __square = square(__size)

    __slash = slash(__size)

    __l1 = l1(__size, __color)
    __l2 = l2(__size, __color)
    __l3 = l3(__size, __color)

    __r1 = r1(__size, __color)
    __r2 = r2(__size, __color)
    __r3 = r3(__size, __color)

    __ps = ps(__size, __color)
    __options = options(__size, __color)
    __share = share(__size, __color)

    @staticmethod
    def __reload():
        PSVG.__up = up(PSVG.__size)
        PSVG.__right = right(PSVG.__size)
        PSVG.__down = down(PSVG.__size)
        PSVG.__left = left(PSVG.__size)

        PSVG.__triangle = triangle(PSVG.__size)
        PSVG.__square = square(PSVG.__size)

        PSVG.__slash = slash(PSVG.__size)

        PSVG.__l1 = l1(PSVG.__size, PSVG.__color)
        PSVG.__l2 = l2(PSVG.__size, PSVG.__color)
        PSVG.__l3 = l3(PSVG.__size, PSVG.__color)

        PSVG.__r1 = r1(PSVG.__size, PSVG.__color)
        PSVG.__r2 = r2(PSVG.__size, PSVG.__color)
        PSVG.__r3 = r3(PSVG.__size, PSVG.__color)

        PSVG.__ps = ps(PSVG.__size, PSVG.__color)
        PSVG.__options = options(PSVG.__size, PSVG.__color)
        PSVG.__share = share(PSVG.__size, PSVG.__color)

    @staticmethod
    def set_color(color: tuple[int, int, int]):
        PSVG.__color = color
        PSVG.__reload()

    @staticmethod
    def set_size(size: int):
        PSVG.__size = size
        PSVG.__reload()

    @staticmethod
    def set_linewidth(linewidth: int):
        PSVG.__linewidth = linewidth

    @staticmethod
    def __outline(surface, center):
        pygame.draw.circle(
            surface,
            PSVG.__color,
            center,
            PSVG.__size,
            PSVG.__linewidth,
        )

    @staticmethod
    def __draw_lines(surface, center, offset_points):
        for i in range(len(offset_points)):
            pygame.draw.line(
                surface,
                PSVG.__color,
                (
                    center[0] + offset_points[i][0],
                    center[1] + offset_points[i][1],
                ),
                (
                    center[0] + offset_points[i - 1][0],
                    center[1] + offset_points[i - 1][1],
                ),
                PSVG.__linewidth,
            )

    @staticmethod
    def up(surface, center):
        PSVG.__outline(surface, center)
        PSVG.__draw_lines(surface, center, PSVG.__up)

    @staticmethod
    def right(surface, center):
        PSVG.__outline(surface, center)
        PSVG.__draw_lines(surface, center, PSVG.__right)

    @staticmethod
    def down(surface, center):
        PSVG.__outline(surface, center)
        PSVG.__draw_lines(surface, center, PSVG.__down)

    @staticmethod
    def left(surface, center):
        PSVG.__outline(surface, center)
        PSVG.__draw_lines(surface, center, PSVG.__left)

    @staticmethod
    def triangle(surface, center):
        PSVG.__outline(surface, center)
        PSVG.__draw_lines(surface, center, PSVG.__triangle)

    @staticmethod
    def circle(surface, center):
        PSVG.__outline(surface, center)
        pygame.draw.circle(
            surface,
            PSVG.__color,
            center,
            PSVG.__size * 0.575,
            PSVG.__linewidth,
        )

    @staticmethod
    def cross(surface, center):
        PSVG.__outline(surface, center)
        pygame.draw.line(
            surface,
            PSVG.__color,
            (
                center[0] + PSVG.__square[0][0],
                center[1] + PSVG.__square[0][1],
            ),
            (
                center[0] + PSVG.__square[2][0],
                center[1] + PSVG.__square[2][1],
            ),
            PSVG.__linewidth,
        )
        pygame.draw.line(
            surface,
            PSVG.__color,
            (
                center[0] + PSVG.__square[1][0],
                center[1] + PSVG.__square[1][1],
            ),
            (
                center[0] + PSVG.__square[3][0],
                center[1] + PSVG.__square[3][1],
            ),
            PSVG.__linewidth,
        )

    @staticmethod
    def square(surface, center):
        PSVG.__outline(surface, center)
        PSVG.__draw_lines(surface, center, PSVG.__square)

    @staticmethod
    def slash(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__slash)

    @staticmethod
    def l1(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__l1[0])
        PSVG.__l1[1].update_pos(center)
        PSVG.__l1[1].draw_to(surface)

    @staticmethod
    def l2(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__l2[0])
        PSVG.__l2[1].update_pos(center)
        PSVG.__l2[1].draw_to(surface)

    @staticmethod
    def l3(surface, center):
        PSVG.__outline(surface, center)
        pygame.draw.circle(
            surface,
            PSVG.__color,
            center,
            PSVG.__size * 0.75,
            PSVG.__linewidth,
        )
        PSVG.__l3.update_pos(center)
        PSVG.__l3.draw_to(surface)

    @staticmethod
    def r1(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__r1[0])
        PSVG.__r1[1].update_pos(center)
        PSVG.__r1[1].draw_to(surface)

    @staticmethod
    def r2(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__r2[0])
        PSVG.__r2[1].update_pos(center)
        PSVG.__r2[1].draw_to(surface)

    @staticmethod
    def r3(surface, center):
        PSVG.__outline(surface, center)
        pygame.draw.circle(
            surface,
            PSVG.__color,
            center,
            PSVG.__size * 0.75,
            PSVG.__linewidth,
        )
        PSVG.__r3.update_pos(center)
        PSVG.__r3.draw_to(surface)

    @staticmethod
    def ps(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__ps[0])
        PSVG.__ps[1].update_pos(center)
        PSVG.__ps[1].draw_to(surface)

    @staticmethod
    def options(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__options[0])
        PSVG.__options[1].update_pos(center)
        PSVG.__options[1].draw_to(surface)

    @staticmethod
    def share(surface, center):
        PSVG.__draw_lines(surface, center, PSVG.__share[0])
        PSVG.__share[1].update_pos(center)
        PSVG.__share[1].draw_to(surface)



import sys

import pygame
from src.classes.PlayStationVectorGraphics.psvg import PSVG

WIN_W, WIN_H = 1100, 500
INDEXES = 9
LINE_INDEXES = 2

screen = pygame.display.set_mode((WIN_W, WIN_H))
fpsclock = pygame.time.Clock()
fps = 60

center = (screen.get_width() // 2, screen.get_height() // 2)


def pos(index: int, line_index: int):
    return (
        int(WIN_W / (INDEXES + 1) * (index + 1)),
        int(WIN_H / (LINE_INDEXES + 1) * (line_index + 1)),
    )


while True:
    events = pygame.event.get()
    for event in events:
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            pygame.quit()
            sys.exit()

    screen.fill((0, 0, 0))

    PSVG.up(screen, pos(0, 0))
    PSVG.right(screen, pos(1, 0))
    PSVG.down(screen, pos(2, 0))
    PSVG.left(screen, pos(3, 0))
    PSVG.slash(screen, pos(4, 0))
    PSVG.triangle(screen, pos(5, 0))
    PSVG.circle(screen, pos(6, 0))
    PSVG.cross(screen, pos(7, 0))
    PSVG.square(screen, pos(8, 0))

    PSVG.l1(screen, pos(0, 1))
    PSVG.l2(screen, pos(1, 1))
    PSVG.l3(screen, pos(2, 1))
    PSVG.r1(screen, pos(3, 1))
    PSVG.r2(screen, pos(4, 1))
    PSVG.r3(screen, pos(5, 1))
    PSVG.ps(screen, pos(6, 1))
    PSVG.options(screen, pos(7, 1))
    PSVG.share(screen, pos(8, 1))

    pygame.display.flip()
    fpsclock.tick(fps)



import pygame

class Table:
    def __init__(
            self, surface, xy: tuple, columns_rows: tuple, x_distance_y_distance: tuple,
            circle_color=(0, 0, 255),
            circle_radius=10):
        """
        xy centers the table at the given position
        """

        self.surface = surface
        self.columns, self.rows = columns_rows[0], columns_rows[1]
        self.x_distance, self.y_distance = x_distance_y_distance[0], x_distance_y_distance[1]
        self.circle_color = circle_color
        self.circle_radius = circle_radius

        x = xy[0]-((self.columns-1)/2)*self.x_distance
        y = xy[1]-((self.rows-1)/2)*self.y_distance
        self.xy = (x, y)

        if self.rows < 1:
            raise ValueError(
                f"Invalid argument for 'rows': '{self.rows}'.")
        if self.columns < 1:
            raise ValueError(
                f"Invalid argument for 'columns': '{self.columns}'.")
        if self.x_distance < 1:
            raise ValueError(
                f"Invalid argument for 'x_distance': '{self.x_distance}'.")
        if self.y_distance < 1:
            raise ValueError(
                f"Invalid argument for 'y_distance': '{self.y_distance}'.")

        # target: dict["row"]["column"] ("self.tdict")
        self.dict = {}
        for x in range(self.columns):
            temp_d = {}
            for y in range(self.rows):
                temp_d[y] = (self.xy[0]+x*self.x_distance,
                             self.xy[1]+y*self.y_distance)
            self.dict[x] = temp_d

    def draw_dots(self):
        for x in range(self.columns):
            for y in range(self.rows):
                pygame.draw.circle(
                    self.surface, self.circle_color, self.dict[x][y], self.circle_radius)

    def get(self, xy: tuple):
        assert isinstance(xy, (tuple, list)), f"invalid argument for 'xy': {xy}"
        assert len(xy) == 2, f"invalid argument for 'xy': {xy}"
        x, y = xy
        try:
            return self.dict[x][y]
        except KeyError:
            raise Exception(f"x or y out of range (x: {x}, y: {y}, min_x: 0, min_y: 0, max_x: {self.columns-1}, max_y: {self.rows-1}")



from .Bar import Bar
from .Button import Button
from .C import C
from .Collisions import Collisions
from .Colors import Colors
from .CustomTemplate import CustomTemplate
from .Entity import Entity
from .Entry import Entry
from .Function import Function
from .Functions import *
from .Keyboard import Keyboard
from .Label import Label
from .Messagebox import Messagebox
from .OneClickManager import OneClickManager
from .Paragraph import Paragraph
from .PerformanceGraph import PerformanceGraph
from .PlayStationController import PlayStationController
from .PlayStationVectorGraphics.psvg import PSVG
from .PopupMessage import PopupMessage
from .RandomClasses import *
from .ScrollableButtonList import ScrollableButtonList
from .Spritesheet import Spritesheet
from .Table import Table



from src.classes.Bar import Bar

import pygame
import sys

c1 = {
    'black': (0,0,0),
    'white': (255,255,255),
    'cyan': (0,255,255),
    'magenta': (255,0,255),
    'yellow': (255,255,0),
    'red': (255,0,0),
    'green': (0,255,0),
    'blue': (0,0,255)
}

pygame.init()
screen = pygame.display.set_mode((500,500))
fpsclock = pygame.time.Clock()
fps = 60

def circle(xy, color, radius=3, width=0):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.circle(screen, color, xy, radius, width)

def line(xy1, xy2, color, width=1):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.line(screen, color, xy1, xy2, width)

def rect(r, color, width=0):
    color = c1[color] if isinstance(color, str) else color
    pygame.draw.rect(screen, color, r, width)

center = (250, 250)

b = Bar(screen, (80,300), center,
           bgc=[80 for i in range(3)],
           fc=(40,40,220),
           bc=(0,0,0),
           borderwidth=3,
           br=15,
           fs="bottom"
           )

val = 50
max_val = 100

while True:
    events = pygame.event.get()
    for event in events:
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            pygame.quit()
            sys.exit()

    mk = pygame.mouse.get_pressed()
    if mk[0]:
        val = max(val - 1, 0)
    if mk[2]:
        val = min(val + 1, max_val)

    if mk[1]:
        b.update_pos(pygame.mouse.get_pos())

    b.update(val, max_val)

    screen.fill((100,100,255))
    b.draw()

    pygame.display.flip()
    fpsclock.tick(fps)


import pygame
from .Label import Label

class Paragraph:
    def __init__(self, surface, text, size, xy: tuple, anchor="center", **kwargs):
        """
        Creates a multiline label (using '\\n').

        Instructions:
            - To create a label, create an instance of this class before
            the mainloop of the game.
            - To make the label appear, call the method 'self.draw()' in
            every loop of the game.

        Example (simplified):
            label = Label(screen, "Hello world!", 32, (100,100), "topleft")
            while True:
                label.draw()

        Arguments:
            surface
                the surface the object will be drawn onto
                Type: pygame.Surface, pygame.display.set_mode
            text
                the text displayed on the label / button
                Type: Any
            size
                refers to the size of text in pixels
                Type: int
            xy
                refers to the position of the anchor
                Type: tuple, list
            anchor
                the anchor of the object:   topleft,    midtop,    topright,
                                            midleft,    center,    midright,
                                            bottomleft, midbottom, bottomright
                Type: str
            textcolor
                color of the text in rgb
                Type: tuple, list
            backgroundcolor
                color of the background in rgb
                Type: None, tuple, list
            antialias
                whether antialias (blurring to make text seem more realistic) should be used
                Type: bool
            font
                the font used when displaying the text; to get all fonts available, use "pygame.font.get_fonts()"
                Type: str
            x_axis_addition
                adds the given amount of pixels to the left and right (-> x axis) of the text
                Type: int
            y_axis_addition
                adds the given amount of pixels to the top and bottom (-> y axis) of the text
                Type: int
            borderwidth
                width of the border
                Type: int
            bordercolor
                color of the border
                Type: tuple, list
            force_width
                forces the width of the background
                Type: int
            force_height
                forces the height of the background
                Type: int
            force_dim
                forces width and height of the background
                Type: tuple, list
            binding_rect
                whether the text_rect or the background_rect should be used with <xy> and <anchor>
                Type: [0, 1]
            borderradius
                used to round the corners of the background_rect; use int for all corners or other types for individual corners
                Type: int, tuple, list
            text_offset
                moves the text by the given values; is inverted if binding_rect == 1
                Type: tuple, list
            image
                blits an image onto the label / button; only works when using <force_dim>
                Type: pygame.Image
            info
                can store any kind of data; usage of dict recommended
                Type: Any
            text_binding
                binds the text within the background_rect
                Type: str (same as <anchor>)
            highlight
                (button only)
                makes the button shine when the mouse hovers over it; when using "True", the button shines a little brighter than
                its usual backgroundcolor; otherwise, the button shines in the given rgb color; only works when using <backgroundcolor>
                Type: True, tuple, list
            active_area
                (button only)
                makes the self.update method only work if the mouse click / hovering is also within the specified rect; useful if
                buttons are moveable but should not be clickable if they are outside a certain area
                Type: tuple, list, pygame.Rect

        All custom arguments can also be used in their short form (eg. "aa" instead of "antialias").
        To see what all the short forms look like, inspect the self.ABBREVIATIONS attribute.
        """

        self.ABBREVIATIONS = {
            "textcolor": "tc",
            "backgroundcolor": "bgc",
            "antialias": "aa",
            "font": "f",
            "x_axis_addition": "xad",
            "y_axis_addition": "yad",
            "borderwidth": "bw",
            "bordercolor": "bc",
            "force_width": "fw",
            "force_height": "fh",
            "force_dim": "fd",
            "binding_rect": "bR",
            "borderradius": "br",
            "text_offset": "to",
            "image": "img",
            "info": "info",
            "text_binding": "tb",
            "highlight": "hl",
            "active_area": "aA",
        }

        self.surface = surface
        self.text = text
        assert type(size) in [int, float], f"invalid argument for 'size': {size}"
        self.size = int(size)
        assert type(xy) in [tuple, list], f"invalid argument for 'xy': {xy}"
        assert len(xy) == 2, f"invalid argument for 'xy': {xy}"
        self.xy = xy
        assert anchor in "topleft,midtop,topright,midleft,center,midright,bottomleft,midbottom,bottomright".split(
            ","), f"invalid argument for 'anchor': {anchor}"
        self.anchor = anchor

        kw = kwargs

        # textcolor
        self.textcolor = kw.get("textcolor", None)
        if self.textcolor == None:
            self.textcolor = kw.get(self.ABBREVIATIONS["textcolor"], None)
        if self.textcolor == None:
            self.textcolor = (255, 255, 255)
        # assertion
        if self.textcolor != None:
            assert type(self.textcolor) in [
                tuple, list], f"invalid argument for 'textcolor': {self.textcolor}"
            assert len(self.textcolor) == 3, f"invalid argument for 'textcolor': {self.textcolor}"

        # backgroundcolor
        self.backgroundcolor = kw.get("backgroundcolor", None)
        if self.backgroundcolor == None:
            self.backgroundcolor = kw.get(self.ABBREVIATIONS["backgroundcolor"], None)
        # assertion
        if self.backgroundcolor != None:
            assert type(self.backgroundcolor) in [
                tuple, list], f"invalid argument for 'backgroundcolor': {self.backgroundcolor}"
            assert len(
                self.backgroundcolor) == 3, f"invalid argument for 'backgroundcolor': {self.backgroundcolor}"

        # backgroundcolor backup
        self.backgroundcolor_init = kw.get("backgroundcolor", None)
        if self.backgroundcolor_init == None:
            self.backgroundcolor_init = kw.get(self.ABBREVIATIONS["backgroundcolor"], None)
        # assertion completed by previous assertion

        # antialias
        self.antialias = kw.get("antialias", None)
        if self.antialias == None:
            self.antialias = kw.get(self.ABBREVIATIONS["antialias"], None)
        if self.antialias == None:
            self.antialias = True
        # assertion
        assert type(self.antialias) == bool, f"invalid argument for 'antialias': {self.antialias}"

        # font
        self.font = kw.get("font", None)
        if self.font == None:
            self.font = kw.get(self.ABBREVIATIONS["font"], None)
        if self.font == None:
            self.font = ""
        # assertion
        if self.font != None:
            assert type(self.font) == str, f"invalid argument for 'font': {self.font}"

        # x_axis_addition
        self.x_axis_addition = kw.get("x_axis_addition", None)
        if self.x_axis_addition == None:
            self.x_axis_addition = kw.get(self.ABBREVIATIONS["x_axis_addition"], None)
        if self.x_axis_addition == None:
            self.x_axis_addition = 0
        # assertion
        if self.x_axis_addition != None:
            assert type(
                self.x_axis_addition) == int, f"invalid argument for 'x_axis_addition': {self.x_axis_addition}"
            assert self.x_axis_addition >= 0, f"invalid argument for 'x_axis_addition': {self.x_axis_addition}"

        # y_axis_addition
        self.y_axis_addition = kw.get("y_axis_addition", None)
        if self.y_axis_addition == None:
            self.y_axis_addition = kw.get(self.ABBREVIATIONS["y_axis_addition"], None)
        if self.y_axis_addition == None:
            self.y_axis_addition = 0
        # assertion
        if self.y_axis_addition != None:
            assert type(
                self.y_axis_addition) == int, f"invalid argument for 'y_axis_addition': {self.y_axis_addition}"
            assert self.y_axis_addition >= 0, f"invalid argument for 'y_axis_addition': {self.y_axis_addition}"

        # borderwidth
        self.borderwidth = kw.get("borderwidth", None)
        if self.borderwidth == None:
            self.borderwidth = kw.get(self.ABBREVIATIONS["borderwidth"], None)
        if self.borderwidth == None:
            self.borderwidth = 0
        # assertion
        if self.borderwidth != None:
            assert type(self.borderwidth) == int, f"invalid argument for 'borderwidth': {self.borderwidth}"
            assert self.borderwidth >= 0, f"invalid argument for 'borderwidth': {self.borderwidth}"

        # bordercolor
        self.bordercolor = kw.get("bordercolor", None)
        if self.bordercolor == None:
            self.bordercolor = kw.get(self.ABBREVIATIONS["bordercolor"], None)
        if self.bordercolor == None:
            self.bordercolor = (0, 0, 0)
        # assertion
        if self.bordercolor != None:
            assert type(self.bordercolor) in [
                tuple, list], f"invalid argument for 'bordercolor': {self.bordercolor}"
            assert len(self.bordercolor) == 3, f"invalid argument for 'bordercolor': {self.bordercolor}"

        # force_width
        self.force_width = kw.get("force_width", None)
        if self.force_width == None:
            self.force_width = kw.get(self.ABBREVIATIONS["force_width"], None)
        # assertion
        if self.force_width != None:
            assert type(self.force_width) == int, f"invalid argument for 'force_width': {self.force_width}"
            assert self.force_width > 0, f"invalid argument for 'force_width': {self.force_width}"

        # force_height
        self.force_height = kw.get("force_height", None)
        if self.force_height == None:
            self.force_height = kw.get(self.ABBREVIATIONS["force_height"], None)
        # assertion
        if self.force_height != None:
            assert type(self.force_height) == int, f"invalid argument for 'force_height': {self.force_height}"
            assert self.force_height > 0, f"invalid argument for 'force_height': {self.force_height}"

        # force_dim
        force_dim = kw.get("force_dim", None)
        if force_dim == None:
            force_dim = kw.get(self.ABBREVIATIONS["force_dim"], None)
        # assertion
        if force_dim != None:
            if type(force_dim) not in [tuple, list] or len(force_dim) != 2:
                raise ValueError(
                    f"Invalid argument for 'force_dim': '{force_dim}'.")
            if force_dim[0] != None:
                self.force_width = force_dim[0]
            if force_dim[1] != None:
                self.force_height = force_dim[1]

        # binding_rect
        self.binding_rect = kw.get("binding_rect", None)
        if self.binding_rect == None:
            self.binding_rect = kw.get(self.ABBREVIATIONS["binding_rect"], None)
        if self.binding_rect == None:
            self.binding_rect = 0
        # assertion
        assert self.binding_rect in [0, 1], f"invalid argument for 'binding_rect': {self.binding_rect}"

        # borderradius
        self.borderradius = kw.get("borderradius", None)
        if self.borderradius == None:
            self.borderradius = kw.get(self.ABBREVIATIONS["borderradius"], None)
        if self.borderradius == None:
            self.borderradius = (0, 0, 0, 0)
        # assertion
        elif type(self.borderradius) == int:
            self.borderradius = (
                self.borderradius, self.borderradius, self.borderradius, self.borderradius)
        elif type(self.borderradius) in [tuple, list]:
            if len(self.borderradius) != 4:
                raise ValueError(
                    f"Invalid argument for 'borderradius': {self.borderradius}.")
        else:
            raise Exception(f"invalid argument for 'borderradius': {self.borderradius}")

        # text_offset
        self.text_offset = kw.get("text_offset", None)
        if self.text_offset == None:
            self.text_offset = kw.get(self.ABBREVIATIONS["text_offset"], None)
        if self.text_offset == None:
            self.text_offset = (0, 0)
        # assertion
        if self.text_offset != None:
            assert type(self.text_offset) in [
                tuple, list], f"invalid argument for 'text_offset': {self.text_offset}"
            assert len(self.text_offset) == 2, f"invalid argument for 'text_offset': {self.text_offset}"

        # image
        self.image = kw.get("image", None)
        if self.image == None:
            self.image = kw.get(self.ABBREVIATIONS["image"], None)
        # # assertion # ! somethings not working here # TODO
        # if self.image != None:
        #     if self.force_width == None or self.force_height == None:
        #         raise ValueError(
        #             "If an image is used, forced dimensions are needed!")
        #     try:
        #         img = pygame.transform.scale(pygame.image.load(self.image), (self.force_width, self.force_height))
        #         self.image = img
        #     except FileNotFoundError:
        #         raise Exception(f"Could not find file '{self.image}'")

        # info
        self.info = kw.get("info", None)
        if self.info == None:
            self.info = kw.get(self.ABBREVIATIONS["info"], None)
        # no assertion needed

        # text_binding
        self.text_binding = kw.get("text_binding", None)
        if self.text_binding == None:
            self.text_binding = kw.get(self.ABBREVIATIONS["text_binding"], None)
        if self.text_binding == None:
            self.text_binding = "center"
        # assertion
        if self.text_binding != None:
            assert self.text_binding in "topleft,midtop,topright,midleft,center,midright,bottomleft,midbottom,bottomright".split(
                ","), f"invalid argument for 'text_binding': {self.text_binding}"

        # highlight
        self.highlight = kw.get("highlight", None)
        if self.highlight == None:
            self.highlight = kw.get(self.ABBREVIATIONS["highlight"], None)
        # assertion
        if self.highlight != None:
            assert type(self.backgroundcolor) in [
                tuple, list], f"'backgroundcolor' (currently: {self.backgroundcolor}) must be defined when using 'highlight' (currently: {self.highlight})"
            if self.highlight == True:
                val = 50
                self.highlight = (
                    min(self.backgroundcolor[0] + val, 255),
                    min(self.backgroundcolor[1] + val, 255),
                    min(self.backgroundcolor[2] + val, 255)
                )
            else:
                assert type(self.highlight) in [
                    tuple, list], f"unknown format for 'highlight': (type: {type(self.highlight)}, value: {self.highlight})"

        # active_area
        self.active_area = kw.get("active_area", None)
        if self.active_area == None:
            self.active_area = kw.get(self.ABBREVIATIONS["active_area"], None)
        # assertion
        if self.active_area != None:
            if type(self.active_area) in [tuple, list]:
                assert len(self.active_area) == 4, f"invalid argument for 'active_area': {self.active_area}"
                self.active_area = pygame.Rect(*self.active_area)
            elif type(self.active_area) == pygame.Rect:
                pass
            else:
                raise Exception(f"invalid argument for 'active_area': {self.active_area}")

        self.kwargs = kwargs
        self.__create__()

    def __create__(self):
        self.__labels__: list[Label] = []
        strings = self.text.split("\n")
        width = 0
        height = 0
        for string in strings:
            l = Label(
                self.surface, string, self.size, self.xy, **self.kwargs
            )
            width = max(width, l.rect.width)
            height = max(height, l.rect.height)

        # removing some arguments by setting them to default
        # bR, fd
        parsed_kwargs = self.kwargs.copy()
        parsed_kwargs["bR"] = 1
        parsed_kwargs["fd"] = (width, height)
        self.__borderwidth__ = self.borderwidth
        parsed_kwargs["bw"] = 0
        self.__borderradius__ = self.borderradius
        parsed_kwargs["br"] = (0, 0, 0, 0)
        self.__bordercolor__ = self.bordercolor
        # kwargs["bc"] = (0, 0, 0) # not needed

        for count, string in enumerate(strings):
            self.__labels__.append(
                Label(
                    self.surface, string, self.size,
                    (self.xy[0],
                     self.xy[1] + height * count),
                    self.anchor, **parsed_kwargs))

        fl = self.__labels__[0]     # first label
        ll = self.__labels__[-1]    # last label
        self.rect = pygame.Rect(fl.rect.left, fl.rect.top, ll.rect.right-fl.rect.left,
                                ll.rect.bottom-fl.rect.top)

    def draw(self):
        for label in self.__labels__:
            label.draw()
        if self.__borderwidth__ > 0:
            pygame.draw.rect(self.surface, self.__bordercolor__, self.rect,
                             self.__borderwidth__, *self.__borderradius__)

    def update_text(self, text: str):
        if str(text) != str(self.text):
            self.text = str(text)
            self.__create__()

    def update_colors(self, textcolor=None, backgroundcolor=None, bordercolor=None):
        if textcolor != None:
            assert type(textcolor) in [
                tuple, list], f"invalid argument for 'textcolor': {textcolor}"
            assert len(textcolor) == 3, f"invalid argument for 'textcolor': {textcolor}"
            self.textcolor = textcolor
            self.__create__()
        if backgroundcolor != None:
            assert type(backgroundcolor) in [
                tuple, list], f"invalid argument for 'backgroundcolor': {backgroundcolor}"
            assert len(
                backgroundcolor) == 3, f"invalid argument for 'backgroundcolor': {backgroundcolor}"
            self.backgroundcolor = backgroundcolor
            self.backgroundcolor_init = backgroundcolor
            val = 50
            self.highlight = (
                min(self.backgroundcolor[0] + val, 255),
                min(self.backgroundcolor[1] + val, 255),
                min(self.backgroundcolor[2] + val, 255)
            )
            self.__create__()
        if bordercolor != None:
            assert type(bordercolor) in [
                tuple, list], f"invalid argument for 'bordercolor': {bordercolor}"
            assert len(bordercolor) == 3, f"invalid argument for 'bordercolor': {bordercolor}"
            self.bordercolor = bordercolor



import pygame
import sys
from .Label import Label
from .Button import Button
from .Functions import vect_sum
from .Colors import Colors
from .Paragraph import Paragraph

class Messagebox:
    def __init__(
            self, surface, pygame_clock: pygame.time.Clock, fps: int, screen_center: tuple, textsize=25,
            textcolor: tuple = (0, 0, 0), backgroundcolor: tuple = (128, 128, 128)):
        self.surface = surface
        self.pygame_clock = pygame_clock
        self.fps = fps
        self.screen_center = screen_center
        self.textsize = textsize
        self.textcolor = textcolor
        self.backgroundcolor = backgroundcolor

        assert fps >= 10
        assert 20 <= textsize <= 40
        assert type(textcolor) in [tuple, list]
        assert type(backgroundcolor) in [tuple, list]

    def askokcancel(self, text, dim_background=True):

        if dim_background:
            self.surface.fill([60 for i in range(3)], special_flags=pygame.BLEND_MULT)

        test = Label(self.surface, text, self.textsize, (0, 0), xad=25)
        width = min(max(test.background_rect.width, 294), 900)
        height = test.rect.height * 5

        rect = pygame.Rect(0, 0, width, height)
        rect.center = self.screen_center

        surface = pygame.Surface((width, height))
        surface.fill(self.backgroundcolor)

        width = min(max(width, 100), 150)
        height = int(test.rect.height * 1.75)
        l_text = Label(self.surface, text, self.textsize, vect_sum(
            self.screen_center, (0, -5)), "midbottom", tc=self.textcolor)

        b_cancel_premade = Button(
            self.surface, "Cancel", self.textsize, vect_sum(rect.midbottom, (3, 0)),
            "bottomright", tc=self.textcolor, bgc=Colors.dark_red, hl=True, bR=1, fd=(width, height),
            bw=4, br=1, bc=(0, 0, 0))
        b_cancel = Button(
            self.surface, "Cancel", self.textsize, vect_sum(rect.midbottom, (3, 0)),
            "bottomright", tc=self.textcolor, bgc=Colors.dark_red, hl=True, bR=1, fd=(width, height),
            bw=4, br=1, bc=(0, 0, 0),
            aA=vect_sum(b_cancel_premade.rect, (3, 3, -6, -6)))

        b_ok_premade = Button(
            self.surface, "OK", self.textsize, vect_sum(rect.midbottom, (-3, 0)),
            "bottomleft", tc=self.textcolor, bgc=Colors.green, hl=True, bR=1, fd=(width, height),
            bw=4, br=1, bc=(0, 0, 0))
        b_ok = Button(
            self.surface, "OK", self.textsize, vect_sum(rect.midbottom, (-3, 0)),
            "bottomleft", tc=self.textcolor, bgc=Colors.green, hl=True, bR=1, fd=(width, height),
            bw=4, br=1, bc=(0, 0, 0),
            aA=vect_sum(b_ok_premade.rect, (3, 3, -6, -6)))

        run = True
        while run:
            event_list = pygame.event.get()
            mouse_events = [event for event in event_list if event.type == pygame.MOUSEBUTTONUP]
            for event in event_list:
                if event.type == pygame.QUIT:
                    pygame.quit()
                    sys.exit()
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        return False

            if b_cancel.update(mouse_events):
                return False
            if b_ok.update(mouse_events):
                return True

            self.surface.blit(surface, rect)
            l_text.draw()
            b_cancel.draw()
            b_ok.draw()

            pygame.draw.rect(self.surface, (0, 0, 0), rect, 4, 1, 1, 1, 1)
            pygame.display.flip()
            self.pygame_clock.tick(self.fps)

    def show_message(self, text, dim_background=True, multiline=False):

        if dim_background:
            self.surface.fill([60 for i in range(3)], special_flags=pygame.BLEND_MULT)

        if multiline:
            test = Paragraph(self.surface, text, self.textsize, (0, 0), xad=25, yad=3)
            width = min(max(test.rect.width, 200), 800)
            height = test.rect.height + 100
        else:
            test = Label(self.surface, text, self.textsize, (0, 0), xad=25)
            width = min(max(test.rect.width, 200), 800)
            height = test.rect.height * 5

        rect = pygame.Rect(0, 0, width, height)
        rect.center = self.screen_center

        surface = pygame.Surface((width, height))
        surface.fill(self.backgroundcolor)

        height = int(test.rect.height * 1.75)
        if multiline:
            lines = len([1 for char in text if char == "\n"])
            jump = self.textsize//2
            l_text = Paragraph(self.surface, text, self.textsize, vect_sum(
                self.screen_center, (0, -lines*jump)), "midbottom", tc=self.textcolor, yad=3)
        else:
            l_text = Label(self.surface, text, self.textsize, vect_sum(
                self.screen_center, (0, -5)), "midbottom", tc=self.textcolor)

        b_ok = Button(
            self.surface, "OK", self.textsize, vect_sum(rect.bottomright, (-10, -10)),
            "bottomright", tc=self.textcolor, bgc=Colors.green, hl=True, bR=1, xad=20, yad=6, to=(0,1),
            bw=4, br=1, bc=(0, 0, 0))

        run = True
        while run:
            event_list = pygame.event.get()
            mouse_events = [event for event in event_list if event.type == pygame.MOUSEBUTTONUP]
            for event in event_list:
                if event.type == pygame.QUIT:
                    pygame.quit()
                    sys.exit()
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        run = False

            if b_ok.update(mouse_events):
                run = False

            self.surface.blit(surface, rect)
            l_text.draw()
            b_ok.draw()

            pygame.draw.rect(self.surface, (0, 0, 0), rect, 4, 1, 1, 1, 1)
            pygame.display.flip()
            self.pygame_clock.tick(self.fps)



import math

class C:
    def __init__(self, *values, mo=False):
        """
        vector class - works very similar to c(...) in the language R
        automatically turns into a list if <mo> (="more_operations") is False,
        if True returns another C-class for further mathematical operations

        NOTE:
        if operations like +=, -=, ... are used, the object will
        always stay a C-class!
        """
        assert len(values) > 0, f"list of values can not be empty"

        if len(values) == 1 and isinstance(values[0], (tuple, list)):
            self.__value = list(values[0])
        else:
            for value in values:
                assert (isinstance(value, (int, float))), \
                    f"values have to be int or float, received: '{value}' (type: {type(value)})"
            self.__value = list(values)
            
        self.__more_operations = bool(mo)

    def __repr__(self):
        return self.__value

    def __round__(self, ndigits=0):
        self.round(ndigits)
        return self

    def round(self, decimals=0):
        assert isinstance(decimals, int), f"invalid argument for 'decimals': {decimals}"
        multiplier = 10 ** decimals
        if decimals == 0:
            for i in range(len(self.__value)):
                self.__value[i] = int(math.floor(self.__value[i]*multiplier + 0.5) / multiplier)
        else:
            for i in range(len(self.__value)):
                self.__value[i] = math.floor(self.__value[i]*multiplier + 0.5) / multiplier

    def __add__(self, other):
        if isinstance(other, C):
            assert len(self.__value) == len(other), f"C-objects must have same length ({len(self.__value)} != {len(other)})"
            if self.__more_operations:
                return C(*[v1+v2 for v1, v2 in zip(self.__value, other)])
            else:
                return [v1+v2 for v1, v2 in zip(self.__value, other)]
        else:
            if self.__more_operations:
                return C(*[v + other for v in self.__value])
            else:
                return [v + other for v in self.__value]

    def __iadd__(self, other):
        self.__value = [v + other for v in self.__value]
        return self

    def __sub__(self, other):
        if isinstance(other, C):
            assert len(self.__value) == len(other), f"C-objects must have same length ({len(self.__value)} != {len(other)})"
            if self.__more_operations:
                return C(*[v1-v2 for v1, v2 in zip(self.__value, other)])
            else:
                return [v1-v2 for v1, v2 in zip(self.__value, other)]
        else:
            if self.__more_operations:
                return C(*[v - other for v in self.__value])
            else:
                return [v - other for v in self.__value]

    def __isub__(self, other):
        self.__value = [v - other for v in self.__value]
        return self

    def __mul__(self, other):
        if isinstance(other, C):
            assert len(self.__value) == len(other), f"C-objects must have same length ({len(self.__value)} != {len(other)})"
            if self.__more_operations:
                return C(*[v1*v2 for v1, v2 in zip(self.__value, other)])
            else:
                return [v1*v2 for v1, v2 in zip(self.__value, other)]
        else:
            if self.__more_operations:
                return C(*[v * other for v in self.__value])
            else:
                return [v * other for v in self.__value]

    def __imul__(self, other):
        self.__value = [v * other for v in self.__value]
        return self

    def __truediv__(self, other):
        if isinstance(other, C):
            assert len(self.__value) == len(other), f"C-objects must have same length ({len(self.__value)} != {len(other)})"
            if self.__more_operations:
                return C(*[v1/v2 for v1, v2 in zip(self.__value, other)])
            else:
                return [v1/v2 for v1, v2 in zip(self.__value, other)]
        else:
            if self.__more_operations:
                return C(*[v / other for v in self.__value])
            else:
                return [v / other for v in self.__value]

    def __itruediv__(self, other):
        self.__value = [v / other for v in self.__value]
        return self

    def __floordiv__(self, other):
        if isinstance(other, C):
            assert len(self.__value) == len(other), f"C-objects must have same length ({len(self.__value)} != {len(other)})"
            if self.__more_operations:
                return C(*[v1//v2 for v1, v2 in zip(self.__value, other)])
            else:
                return [v1//v2 for v1, v2 in zip(self.__value, other)]
        else:
            if self.__more_operations:
                return C(*[v // other for v in self.__value])
            else:
                return [v // other for v in self.__value]

    def __ifloordiv__(self, other):
        self.__value = [v // other for v in self.__value]
        return self

    def __iter__(self):
        self.__count = 0
        self.__max = len(self.__value)
        return self

    def __next__(self):
        if self.__count < self.__max:
            v = self.__value[self.__count]
            self.__count += 1
            return v
        else:
            raise StopIteration

    def __len__(self):
        return len(self.__value)

    def __getitem__(self, key):
        return self.__value[key]

    def value(self):
        return self.__value


import pygame
import os
import json

class Spritesheet:
    def __init__(self, path_to_folder: str, images_facing: str = "right"):
        """
        <name> must be the path to the folder containing the files listed below \n
        every file inside that folder must have the same name as the folder itself (ignoring endings) \n

        use self.get_frames to get images from the spritesheet (not self.__get_frame__ !)

        files necessary:
            - <name>.png
            - <name>.json (containing metadata)

        (works only with aseprite, best with horizontal strip)

        the json file has to have the following structure (the names
        of the frames have to be called just by their number):
        {
            "frames": {
                1: {
                    "frame": {
                        "x": ...,
                        "y": ...,
                        "w": ...,
                        "h": ...
                    }
                }
            }
        }
        """
        foldername = path_to_folder.split(os.sep)[-1]
        png_file = os.path.join(os.path.abspath(path_to_folder), (foldername + ".png"))
        json_file = os.path.join(os.path.abspath(path_to_folder), (foldername + ".json"))
        if not os.path.exists(png_file):
            raise FileNotFoundError(f"Could not find file {png_file}")
        elif not os.path.exists(json_file):
            raise FileNotFoundError(f"Could not find file {json_file}")

        self.base_image = pygame.image.load(png_file)
        with open(json_file) as f:
            self.metadata = json.load(f)

        assert images_facing in ["left", "right"], "invalid direction"
        self.images_facing = images_facing

    def __get_frame__(
            self, frame_number: int, colorkey: tuple[int, int, int] = (0, 0, 0),
            resize_factor: int = 1) -> dict:
        """returns a dict with images (left, right, flipped_left, flipped_right)"""

        assert self.metadata["frames"][str(frame_number)], f"no frame with number '{frame_number}' found"
        assert resize_factor > 0, "resize factor must be above 0"

        x = self.metadata["frames"][str(frame_number)]["frame"]["x"]
        y = self.metadata["frames"][str(frame_number)]["frame"]["y"]
        width = self.metadata["frames"][str(frame_number)]["frame"]["w"]
        height = self.metadata["frames"][str(frame_number)]["frame"]["h"]
        base_surface = pygame.Surface((width, height))
        base_surface.set_colorkey(colorkey)
        base_surface.blit(self.base_image, (0, 0), (x, y, width, height))

        if resize_factor != 1:
            new_width, new_height = int(width*resize_factor), int(height*resize_factor)
            base_surface = pygame.transform.scale(base_surface, (new_width, new_height))

        if self.images_facing == "right":
            img_dict = {
                "right": base_surface,
                "left": pygame.transform.flip(base_surface, True, False),
                "flipped_right": pygame.transform.flip(base_surface, False, True),
                "flipped_left": pygame.transform.flip(base_surface, False, True)
            }
        elif self.images_facing == "left":
            img_dict = {
                "right": pygame.transform.flip(base_surface, True, False),
                "left": base_surface,
                "flipped_right": pygame.transform.flip(base_surface, True, True),
                "flipped_left": pygame.transform.flip(base_surface, False, True)
            }

        return img_dict

    def get_frames(self, first_frame_number: int, last_frame_number: int,
                   colorkey: tuple[int, int, int] = (0, 0, 0),
                   resize_factor: int = 1) -> list[dict]:
        """ returns a list of the requested frames \n
        first_frame_number and last_frame_number are inclusive """
        try:
            return [self.__get_frame__(i, colorkey, resize_factor) for i in range(first_frame_number, last_frame_number + 1)]
        except KeyError:
            raise KeyError(
                f"key not found. you probably requested more frames than there are pictures in this spritesheet")

    def get_size(self, frame_number=0):
        width = self.metadata["frames"][str(frame_number)]["frame"]["w"]
        height = self.metadata["frames"][str(frame_number)]["frame"]["h"]
        return [width, height]
